#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>

#include <gtk/gtk.h>

#include "gtklife.h"
#include "life.h"
#include "util.h"
#include "icons.h"

/* Constants */

static GtkItemFactoryEntry menu_items[] = {
    { "/_File",                  0,                 0,                      0,  "<Branch>"     },
    { "/File/tearoff",           0,                 0,                      0,  "<Tearoff>"    },
    { "/File/_New",              "<control>N",      file_new,               0,  0              },
    { "/File/_Open...",          "<control>O",      file_open,              0,  0              },
    { "/File/_Reopen...",        "R",               file_reopen,            0,  0              },
    { "/File/_Save",             "<control>S",      file_save,              0,  0              },
    { "/File/Save _As...",       "<control>A",      file_save_as,           0,  0              },
    { "/File/sep1",              0,                 0,                      0,  "<Separator>"  },
    { "/File/_Description...",   "D",               file_description,       0,  0              },
    { "/File/sep2",              0,                 0,                      0,  "<Separator>"  },
    { "/File/_Quit",             "<control>Q",      file_quit,              0,  0              },

    { "/_View",                  0,                 0,                      0,  "<Branch>"     },
    { "/View/tearoff",           0,                 0,                      0,  "<Tearoff>"    },
    { "/View/Zoom _In",          "equal",           view_zoom_in,           0,  0              },
    { "/View/Zoom _Out",         "minus",           view_zoom_out,          0,  0              },
    { "/View/_Zoom",             0,                 0,                      0,  "<Branch>"     },
    { "/View/Zoom/tearoff",      0,                 0,                      0,  "<Tearoff>"    },
    { "/View/Zoom/_1:1",         "1",               view_zoom_1,            0,  0              },
    { "/View/Zoom/_2:1",         "2",               view_zoom_2,            0,  0              },
    { "/View/Zoom/_4:1",         "4",               view_zoom_4,            0,  0              },
    { "/View/Zoom/_8:1",         "8",               view_zoom_8,            0,  0              },
    { "/View/Zoom/1_6:1",        "Z",               view_zoom_16,           0,  0              },
    { "/View/sep1",              0,                 0,                      0,  "<Separator>"  },
    { "/View/Show _Toolbar",     0,                 view_show_toolbar,      0,  "<CheckItem>"  },
    { "/View/Show _Sidebar",     0,                 view_show_sidebar,      0,  "<CheckItem>"  },
    { "/View/Show S_crollbars",  0,                 view_show_scrollbars,   0,  "<CheckItem>"  },

    /* Invisible submenu, just used to set accelerators */
    { "/_Position",              0,                 0,                      0,  "<Branch>"     },
    { "/Position/Center",        "KP_5",            position_center,        0,  0              },
    { "/Position/Scroll Left",   "KP_4",            position_scroll_left,   0,  0              },
    { "/Position/Scroll Right",  "KP_6",            position_scroll_right,  0,  0              },
    { "/Position/Scroll Up",     "KP_8",            position_scroll_up,     0,  0              },
    { "/Position/Scroll Down",   "KP_2",            position_scroll_down,   0,  0              },
    { "/Position/Scroll NW",     "KP_7",            position_scroll_nw,     0,  0              },
    { "/Position/Scroll NE",     "KP_9",            position_scroll_ne,     0,  0              },
    { "/Position/Scroll SW",     "KP_1",            position_scroll_sw,     0,  0              },
    { "/Position/Scroll SE",     "KP_3",            position_scroll_se,     0,  0              },
    { "/Position/Page Left",     "<control>KP_4",   position_page_left,     0,  0              },
    { "/Position/Page Right",    "<control>KP_6",   position_page_right,    0,  0              },
    { "/Position/Page Up",       "<control>KP_8",   position_page_up,       0,  0              },
    { "/Position/Page Down",     "<control>KP_2",   position_page_down,     0,  0              },
    { "/Position/Page NW",       "<control>KP_7",   position_page_nw,       0,  0              },
    { "/Position/Page NE",       "<control>KP_9",   position_page_ne,       0,  0              },
    { "/Position/Page SW",       "<control>KP_1",   position_page_sw,       0,  0              },
    { "/Position/Page SE",       "<control>KP_3",   position_page_se,       0,  0              },

    { "/_Edit",                  0,                 0,                      0,  "<Branch>"     },
    { "/Edit/tearoff",           0,                 0,                      0,  "<Tearoff>"    },
    { "/Edit/_Preferences",      0,                 edit_preferences,       0,  0              },

    { "/_Run",                   0,                 0,                      0,  "<Branch>"     },
    { "/Run/tearoff",            0,                 0,                      0,  "<Tearoff>"    },
    { "/Run/_Start",             "S",               run_start_stop,         0,  0              },
    { "/Run/S_tep",              "T",               run_step,               0,  0              },
    { "/Run/sep1",               0,                 0,                      0,  "<Separator>"  },
    { "/Run/_Faster",            "period",          run_faster,             0,  0              },
    { "/Run/Slo_wer",            "comma",           run_slower,             0,  0              },
    { "/Run/S_peed...",          0,                 run_speed,              0,  0              },

    { "/_Help",                  0,                 0,                      0,  "<LastBranch>" },
    { "/Help/tearoff",           0,                 0,                      0,  "<Tearoff>"    },
    { "/Help/_Help...",          "H",               help_help,              0,  0              },
    { "/Help/_About...",         0,                 help_about,             0,  0              }
};

static toolbar_item toolbar_items[] = {
    { "Clear the grid",                 new_icon,          file_new         },
    { "Open a pattern file",            open_icon,         file_open        },
    { "Save pattern",                   save_icon,         file_save        },
    { "Save pattern as",                save_as_icon,      file_save_as     },
    { "View/edit pattern description",  description_icon,  file_description },

    {"sep",NULL,NULL},

    { "Zoom in",                        zoom_in_icon,      view_zoom_in     },
    { "Zoom out",                       zoom_out_icon,     view_zoom_out    },
    { "Zoom 1:1",                       zoom_1_icon,       view_zoom_1      },
    { "Zoom all the way in",            zoom_16_icon,      view_zoom_16     },

    {"sep",NULL,NULL},

    { "Edit preferences",               preferences_icon,  edit_preferences },

    {"sep",NULL,NULL},

    { "Start/stop",                     start_icon,        run_start_stop   },
    { "Step ahead one generation",      step_icon,         run_step         },

    {NULL,NULL,NULL}
};

/* Configurable Globals */

static char*        patterns_dir            = DEFAULT_PATTERNS_DIR;
static int32        init_window_width       = DEFAULT_WINDOW_WIDTH;
static int32        init_window_height      = DEFAULT_WINDOW_HEIGHT;
static int32        default_zoom            = DEFAULT_ZOOM;
static boolean      default_show_toolbar    = DEFAULT_SHOW_TOOLBAR;
static boolean      default_show_sidebar    = DEFAULT_SHOW_SIDEBAR;
static boolean      default_show_scrollbars = DEFAULT_SHOW_SCROLLBARS;
static int32        default_speed           = DEFAULT_SPEED;
static int32        speed_slider_min        = DEFAULT_SPEED_SLIDER_MIN;
static int32        speed_slider_max        = DEFAULT_SPEED_SLIDER_MAX;
static int32        speed_increment         = DEFAULT_SPEED_INCREMENT;
static boolean      default_skip_frames     = DEFAULT_SKIP_FRAMES;
static char*        help_browser            = DEFAULT_HELP_BROWSER;

static uint32       colors[] = {DEFAULT_BG_COLOR, DEFAULT_CELL_COLOR, DEFAULT_GRID_COLOR};

/* Other Globals */

static char*        user_dir                = NULL;
static char*        current_dir            = NULL;
static char*        pattern_path            = NULL;
static uint8*       life_pixmap             = NULL;
static uint32       life_pixmap_alloc_len   = 0;
static GdkRgbCmap*  life_pixmap_colormap;
static boolean      pattern_running         = FALSE;
static int32        zoom;
static int32        speed;
static boolean      skip_frames;
static uint32       start_tick;
static uint64       start_time;
static int32        skipped_frames;
static boolean      show_toolbar, show_sidebar, show_scrollbars;

static recent_file  recent_files[MAX_RECENT_FILES];
static GtkWidget*   recent_files_separator;

/* Canvas is measured in pixels, viewport in cells */
static int32        canvas_width, canvas_height;
static int32        viewport_widths[MAX_ZOOM+1], viewport_heights[MAX_ZOOM+1];
static int32        viewport_width, viewport_height;
static int32        viewport_xstart, viewport_ystart, viewport_xend, viewport_yend;

/* Clipping rectangle, in pixels */
static int32        update_xstart, update_ystart, update_xend, update_yend;

static boolean      dragging_mouse = FALSE;
static int32        last_x, last_y;

static GtkWidget*   window;
static GtkWidget*   toolbar_box;
static GtkWidget*   sidebar;
static GtkWidget*   pattern_list;
static GtkWidget*   canvas;
static GtkWidget*   vscrollbar;
static GtkWidget*   hscrollbar;
static GtkWidget*   start_stop_menu_label;
static GtkWidget*   start_stop_toolbar_button;
static GtkWidget*   show_toolbar_menu_item;
static GtkWidget*   show_sidebar_menu_item;
static GtkWidget*   show_scrollbars_menu_item;
static GtkWidget*   start_pixmap;
static GtkWidget*   stop_pixmap;
static GtkWidget*   speed_label;
static GtkWidget*   speed_slider;
static GtkWidget*   tick_label;
static GtkWidget*   population_label;

static GtkItemFactory*  menu_factory;

/* Main & basic initialization functions */

int main(int argc, char *argv[]) {
    set_prog_name(PROG);
    create_user_dir();
    load_preferences();
    apply_defaults();
    load_recent_files_list();
    gtk_init(&argc, &argv);
    init_gui();
    load_keybindings();
    if (argc > 1)
        attempt_load_pattern(argv[1]);
    main_loop();

    return 0;
}

/* Functions for reading/saving user settings */

/* Create the user directory (if necessary), and place the full user dir path in the global
 * user_dir.
 */
void create_user_dir(void)
{
    const char*  home_dir;

    home_dir = getenv("HOME");
    if (!home_dir) {
        warn("No home directory found! Setting user dir to '/'");
        user_dir = "";
        return;
    }
    user_dir = dsprintf("%s/%s", home_dir, USER_DIR);
    if (mkdir(user_dir, 0777) < 0 && errno != EEXIST) {
        sys_warn("can't create user directory %s", user_dir);
        return;
    }
}

/* Read and apply saved user preferences
 */
void load_preferences(void)
{
    FILE*  f;
    char*  path;
    char*  line;
    char*  key;
    char*  val;
    char*  equal_pos;

    patterns_dir = safe_strdup(patterns_dir);
    help_browser = safe_strdup(help_browser);

    path = dsprintf("%s/preferences", user_dir);
    f = fopen(path, "r");
    free(path);
    if (!f)
        return;

    while ((line = dgets(f)) != NULL) {
        equal_pos = strchr(line, '=');
        if (!equal_pos) {
            warn("invalid line found in preferences file");
            free(line);
            continue;
        }
        *equal_pos = '\0';
        key = line;
        val = equal_pos+1;

        if (STR_EQUAL(key, "default_patterns_dir")) {
            free(patterns_dir);
            patterns_dir = safe_strdup(val);
        } else if (STR_EQUAL(key, "init_window_width"))
            init_window_width = atoi(val);
        else if (STR_EQUAL(key, "init_window_height"))
            init_window_height = atoi(val);
        else if (STR_EQUAL(key, "default_zoom"))
            default_zoom = atoi(val);
        else if (STR_EQUAL(key, "default_show_toolbar"))
            default_show_toolbar = atoi(val);
        else if (STR_EQUAL(key, "default_show_sidebar"))
            default_show_sidebar = atoi(val);
        else if (STR_EQUAL(key, "default_show_scrollbars"))
            default_show_scrollbars = atoi(val);
        else if (STR_STARTS_WITH(key, "color_") && is_numeric(key+6) && atoi(key+6) < NUM_COLORS)
            colors[atoi(key+6)] = strtoul(val, NULL, 16);
        else if (STR_EQUAL(key, "default_speed"))
            default_speed = atoi(val);
        else if (STR_EQUAL(key, "speed_slider_min"))
            speed_slider_min = atoi(val);
        else if (STR_EQUAL(key, "speed_slider_max"))
            speed_slider_max = atoi(val);
        else if (STR_EQUAL(key, "speed_increment"))
            speed_increment = atoi(val);
        else if (STR_EQUAL(key, "default_skip_frames"))
            default_skip_frames = atoi(val);
        else if (STR_EQUAL(key, "help_browser")) {
            free(help_browser);
            help_browser = safe_strdup(val);
        } else
            warn("invalid key '%s' found in preferences file", key);

        free(line);
    }

    fclose(f);
}

/* Apply default values to various globals, based on preferences and/or system defaults.
 */
void apply_defaults(void)
{
    zoom            = default_zoom;
    show_toolbar    = default_show_toolbar;
    show_sidebar    = default_show_sidebar;
    show_scrollbars = default_show_scrollbars;
    speed           = default_speed;
    skip_frames     = default_skip_frames;
    current_dir     = safe_strdup(patterns_dir);
}

/* Read and apply saved user keybindings
 */
void load_keybindings(void)
{
    char*  path;

    path = dsprintf("%s/keybindings", user_dir);
    gtk_item_factory_parse_rc(path);
    free(path);
}

/* Read the recent file list to put in the "File" menu
 */
void load_recent_files_list(void)
{
    FILE*  f;
    char*  rf_path;
    char*  full_path;
    char*  filename;
    char*  slash_pos;
    int32  i;

    memset(recent_files, 0, sizeof(recent_file) * MAX_RECENT_FILES);
    rf_path = dsprintf("%s/recent", user_dir);
    f = fopen(rf_path, "r");
    free(rf_path);
    if (!f)
        return;
    i = 0;
    while (i < MAX_RECENT_FILES && (full_path = dgets(f)) != NULL) {
        slash_pos = strrchr(full_path, '/');
        filename = safe_strdup(slash_pos ? slash_pos+1 : full_path);
        recent_files[i].full_path = full_path;
        recent_files[i].filename  = filename;
        i++;
    }
    fclose(f);
}

/* Save current preferences to a file in the user directory.
 */
void save_preferences(void)
{
    FILE*  f;
    char*  path;
    int32  i;

    path = dsprintf("%s/preferences", user_dir);
    f = fopen(path, "w");
    free(path);
    if (!f) {
        sys_warn("can't open %s/preferences for writing", user_dir);
        return;
    }

    fprintf(f, "default_patterns_dir=%s\n",     patterns_dir);
    fprintf(f, "init_window_width=%d\n",        init_window_width);
    fprintf(f, "init_window_height=%d\n",       init_window_height);
    fprintf(f, "default_zoom=%d\n",             default_zoom);
    fprintf(f, "default_show_toolbar=%d\n",     default_show_toolbar);
    fprintf(f, "default_show_sidebar=%d\n",     default_show_sidebar);
    fprintf(f, "default_show_scrollbars=%d\n",  default_show_scrollbars);
    for (i=0; i < NUM_COLORS; i++)
        fprintf(f, "color_%d=%06X\n", i, colors[i]);
    fprintf(f, "default_speed=%d\n",            default_speed);
    fprintf(f, "speed_slider_min=%d\n",         speed_slider_min);
    fprintf(f, "speed_slider_max=%d\n",         speed_slider_max);
    fprintf(f, "speed_increment=%d\n",          speed_increment);
    fprintf(f, "default_skip_frames=%d\n",      default_skip_frames);
    fprintf(f, "help_browser=%s\n",             help_browser);

    if (fclose(f) != 0)
        sys_warn("close failed after writing to ~/%s/preferences", USER_DIR);
}

/* Save current keybindings to a file in the user directory.
 */
void save_keybindings(void)
{
    char*  path;

    path = dsprintf("%s/keybindings", user_dir);
    gtk_item_factory_dump_rc(path, NULL, TRUE);
    free(path);
}

/* Save the list of recently accessed pattern files
 */
void save_recent_files_list(void)
{
    FILE*  f;
    char*  path;
    int32  i;

    path = dsprintf("%s/recent", user_dir);
    f = fopen(path, "w");
    free(path);
    if (!f) {
        sys_warn("can't open %s/recent for writing", user_dir);
        return;
    }
    for (i=0; i < MAX_RECENT_FILES && recent_files[i].full_path; i++)
        fprintf(f, "%s\n", recent_files[i].full_path);
    if (fclose(f) != 0)
        sys_warn("close failed after writing to ~/%s/recent", USER_DIR);
}

/* Main loop functions */

/* Handle events and run the Life pattern
 */
void main_loop(void)
{
    boolean block;

    while (1) {
        /* Block if no pattern is running and they're not dragging the mouse, else poll */
        while ((block = !pattern_running && !dragging_mouse) || gtk_events_pending())
            gtk_main_iteration_do(block);
        if (dragging_mouse)
            handle_mouse_drag();
        if (pattern_running)
            tick_and_update();
        else
            usleep(1);   /* don't eat up all the processor just for mouse dragging */
    }
}

/* If we're in sync with the requested speed (or the pattern is being stepped through manually),
 * process the next Life tick, update the pixmap and redraw the screen. If going too fast, don't do
 * anything. If too slow and skip_frames is TRUE, and skipped_frames is not maxed out yet, process
 * the next tick and update the pixmap, but don't redraw the screen.
 */
void tick_and_update(void)
{
    int32  running_fps = 0;

    if (pattern_running) {
        running_fps = ROUND((double)(tick - start_tick) * 1000.0 /
                            (double)(get_time_milliseconds() - start_time));
        if (running_fps > speed)
            return;
    }
    if (!skipped_frames) {
        update_xstart = canvas_width  - 1;
        update_ystart = canvas_height - 1;
        update_xend = update_yend = 0;
    }

    next_tick();

    if (pattern_running && running_fps < speed && skip_frames &&
        skipped_frames < MAX_SKIPPED_FRAMES)
        skipped_frames++;
    else {
        trigger_canvas_update();
        update_tick_label();
        update_population_label();
    }
}

/* GUI initialization functions */

/* Setup the gtklife window and widgets
 */
void init_gui(void)
{
    GtkWidget*  vbox;
    GtkWidget*  hbox;
    int32       i;

    init_rgb();
    init_window();

    vbox = gtk_vbox_new(FALSE, 0);
    init_menubar(vbox);
    init_toolbar(vbox);
    hbox = gtk_hbox_new(FALSE, 0);
    init_sidebar(hbox);
    init_canvas(hbox);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
    gtk_container_add(GTK_CONTAINER(window), vbox);
    gtk_widget_show_all(window);

    /* Hide the "Position" submenu -- we only use it to set accelerators */
    gtk_widget_hide(gtk_item_factory_get_item(menu_factory, "/Position"));

    /* Hide any GUI components the user doesn't want to see */
    if (!show_toolbar)
        gtk_widget_hide_all(toolbar_box);
    if (!show_sidebar)
        gtk_widget_hide_all(sidebar);
    if (!show_scrollbars) {
        gtk_widget_hide(vscrollbar);
        gtk_widget_hide(hscrollbar);
    }

    /* Hide unused recent-file entries. If there are no entries, hide the separator too. */
    for (i=0; i < MAX_RECENT_FILES; i++) {
        if (!recent_files[i].filename)
            gtk_widget_hide(recent_files[i].menu_item);
    }
    if (!recent_files[0].filename)
        gtk_widget_hide(recent_files_separator);

    /* Put keyboard focus on the speed slider */
    gtk_widget_grab_focus(speed_slider);
}

/* Initialize the GdkRGB subsystem for image handling. Also create the colormap we'll be using.
 */
void init_rgb(void)
{
    gdk_rgb_init();
    gtk_widget_set_default_colormap(gdk_rgb_get_cmap());
    gtk_widget_set_default_visual(gdk_rgb_get_visual());
    life_pixmap_colormap = gdk_rgb_cmap_new(colors, NUM_COLORS);
}

/* Initialize the main application window
 */
void init_window(void)
{
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(file_quit), "WM Destroy");
    gtk_widget_set_name(GTK_WIDGET(window), PROG);
    gtk_window_set_title(GTK_WINDOW(window), TITLE);
    gtk_window_set_default_size(GTK_WINDOW(window), init_window_width, init_window_height);
    gtk_widget_realize(window);
}

/* Setup the top menubar
 */
void init_menubar(GtkWidget* containing_box)
{
    GtkWidget*       menubar;
    GtkAccelGroup*   accelerators;
    GtkWidget*       file_menu;
    GtkWidget*       start_stop_menu_item;
    GtkWidget*       item;
    char*            str;
    int32            num_items;
    int32*           num;
    int32            i;

    /* Setup menu items */
    accelerators = gtk_accel_group_new();
    menu_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accelerators);
    num_items = sizeof(menu_items) / sizeof(menu_items[0]);
    gtk_item_factory_create_items(menu_factory, num_items, menu_items, NULL);
    gtk_window_add_accel_group(GTK_WINDOW(window), accelerators);
    menubar = gtk_item_factory_get_widget(menu_factory, "<main>");
    start_stop_menu_item = gtk_item_factory_get_item(menu_factory, "/Run/Start");
    start_stop_menu_label = GTK_BIN(start_stop_menu_item)->child;
    show_toolbar_menu_item = gtk_item_factory_get_item(menu_factory, "/View/Show Toolbar");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(show_toolbar_menu_item), show_toolbar);
    show_sidebar_menu_item = gtk_item_factory_get_item(menu_factory, "/View/Show Sidebar");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(show_sidebar_menu_item), show_sidebar);
    show_scrollbars_menu_item = gtk_item_factory_get_item(menu_factory, "/View/Show Scrollbars");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(show_scrollbars_menu_item),
                                   show_scrollbars);

    /* Setup recent files list */
    file_menu = gtk_item_factory_get_widget(menu_factory, "/File");
    for (i=0; i < MAX_RECENT_FILES; i++) {
        if (recent_files[i].filename)
            str = dsprintf("%d. %s", i+1, recent_files[i].filename);
        else
            str = dsprintf("%d.", i+1);
        item = gtk_menu_item_new_with_label(str);
        free(str);
        recent_files[i].menu_item = item;
        recent_files[i].label     = GTK_BIN(item)->child;
        gtk_label_set_pattern(GTK_LABEL(recent_files[i].label), "_");
        gtk_widget_add_accelerator(item, "activate_item",
                                   gtk_menu_get_uline_accel_group(GTK_MENU(file_menu)),
                                   (i+1) + 0x30, 0, GTK_ACCEL_LOCKED);
        gtk_widget_add_accelerator(item, "activate_item", accelerators, (i+1) + 0x30,
                                   GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
        num = safe_malloc(sizeof(int32));
        *num = i;
        gtk_signal_connect(GTK_OBJECT(item), "activate", GTK_SIGNAL_FUNC(file_recent), num);
        gtk_menu_shell_insert(GTK_MENU_SHELL(file_menu), item, RECENT_FILES_INSERT_POS+i);
    }
    /* Insert separator after recent files */
    recent_files_separator = gtk_menu_item_new();
    gtk_menu_shell_insert(GTK_MENU_SHELL(file_menu), recent_files_separator,
                          RECENT_FILES_INSERT_POS+i);

    /* Add menubar to the window */
    gtk_box_pack_start(GTK_BOX(containing_box), menubar, FALSE, FALSE, 0);
}

/* Setup the toolbar, including speed slider and tick display to the right of the buttons
 */
void init_toolbar(GtkWidget* containing_box)
{
    GtkWidget*     toolbar;
    GtkWidget*     icon;
    GtkWidget*     button;
    GtkWidget*     toolbar_frame;
    GtkWidget*     speed_frame;
    GtkWidget*     tick_and_pop_frame;
    GtkWidget*     hbox;
    toolbar_item*  item;

    /* Create the button bar */
    toolbar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);
    gtk_toolbar_set_space_size(GTK_TOOLBAR(toolbar), 10);

    /* Add toolbar buttons */
    for (item=toolbar_items; item->description; item++) {
        if (STR_EQUAL(item->description, "sep"))
            gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));
        else {
            icon = load_pixmap_from_rgba_array(item->icon, ICON_WIDTH, ICON_HEIGHT,
                                               &(window->style->bg[GTK_STATE_NORMAL]));
            button = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), NULL, item->description, NULL,
                                             icon, item->callback, NULL);
            if (item->icon == start_icon) {
                start_stop_toolbar_button = button;
                start_pixmap = icon;
            }
        }
    }
    stop_pixmap = load_pixmap_from_rgba_array(stop_icon, ICON_WIDTH, ICON_HEIGHT,
                                               &(window->style->bg[GTK_STATE_NORMAL]));

    /* Frame the button bar */
    toolbar_frame = gtk_frame_new(NULL);
    gtk_container_add(GTK_CONTAINER(toolbar_frame), toolbar);

    /* Create and frame the Gen/s slider */
    hbox = gtk_hbox_new(FALSE, 0);
    init_speed_box(hbox);
    speed_frame = gtk_frame_new(NULL);
    gtk_container_add(GTK_CONTAINER(speed_frame), hbox);

    /* Create and frame the tick and population indicators */
    hbox = gtk_hbox_new(FALSE, 0);
    init_tick_display(hbox);
    init_population_display(hbox);
    tick_and_pop_frame = gtk_frame_new(NULL);
    gtk_container_add(GTK_CONTAINER(tick_and_pop_frame), hbox);

    /* Pack the button bar, speed control, and tick display into toolbar_box */
    toolbar_box = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(toolbar_box), toolbar_frame, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(toolbar_box), speed_frame, TRUE, TRUE, 10);
    gtk_box_pack_end(GTK_BOX(toolbar_box), tick_and_pop_frame, FALSE, FALSE, 0);

    /* Add the toolbar box to the container */
    gtk_box_pack_start(GTK_BOX(containing_box), toolbar_box, FALSE, FALSE, 0);
}

/* Setup the speed slider box
 */
void init_speed_box(GtkWidget* containing_box)
{
    GtkWidget*  label;
    GtkObject*  slider_params;
    char*       speed_str;

    /* Setup label */
    label = gtk_label_new(" Speed:  ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);
    speed_str = dsprintf("%d", speed);
    speed_label = gtk_label_new(speed_str);
    free(speed_str);
    set_speed_label_size(speed_slider_max);
    gtk_misc_set_alignment(GTK_MISC(speed_label), 1, 0.5);
    gtk_box_pack_start(GTK_BOX(containing_box), speed_label, FALSE, FALSE, 0);
    label = gtk_label_new(" Gen/s  ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);

    /* Setup slider */
    slider_params = gtk_adjustment_new(speed, speed_slider_min, speed_slider_max, 1,
                                       speed_increment, 0);
    speed_slider = gtk_hscale_new(GTK_ADJUSTMENT(slider_params));
    gtk_scale_set_draw_value(GTK_SCALE(speed_slider), FALSE);
    gtk_signal_connect(GTK_OBJECT(slider_params), "value_changed",
                       GTK_SIGNAL_FUNC(handle_speed_slider_change), NULL);
    gtk_box_pack_start(GTK_BOX(containing_box), speed_slider, TRUE, TRUE, 0);

    /* Extra space at the end */
    label = gtk_label_new("  ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);
}

/* Set up the tick display
 */
void init_tick_display(GtkWidget* containing_box)
{
    GtkWidget*  label;

    label = gtk_label_new(" Tick:  ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);
    tick_label = gtk_label_new("0");
    gtk_box_pack_start(GTK_BOX(containing_box), tick_label, FALSE, FALSE, 0);
    label = gtk_label_new("   ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);
}

/* Set up the population display
 */
void init_population_display(GtkWidget* containing_box)
{
    GtkWidget*  label;

    label = gtk_label_new("Pop:  ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);
    population_label = gtk_label_new("0");
    gtk_box_pack_start(GTK_BOX(containing_box), population_label, FALSE, FALSE, 0);
    label = gtk_label_new("  ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);
}

/* Setup the pattern loader sidebar
 */
void init_sidebar(GtkWidget* containing_box)
{
    char*  header = "Patterns";

    pattern_list = gtk_clist_new_with_titles(1, &header);
    update_pattern_list();
    gtk_clist_column_titles_passive(GTK_CLIST(pattern_list));
    gtk_clist_set_selection_mode(GTK_CLIST(pattern_list), GTK_SELECTION_SINGLE);
    gtk_signal_connect(GTK_OBJECT(pattern_list), "select_row",
                       GTK_SIGNAL_FUNC(handle_sidebar_select), NULL);
    sidebar = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sidebar), GTK_POLICY_AUTOMATIC,
                                   GTK_POLICY_AUTOMATIC);
    gtk_container_add(GTK_CONTAINER(sidebar), pattern_list);
    gtk_box_pack_start(GTK_BOX(containing_box), sidebar, FALSE, FALSE, 0);
}

/* Setup the Life pattern drawing area
 */
void init_canvas(GtkWidget* containing_box)
{
    GtkWidget*  vbox;
    GtkWidget*  hbox;
    GtkObject*  scroll_params;

    vbox = gtk_vbox_new(FALSE, 0);
    hbox = gtk_hbox_new(FALSE, 0);

    /* Create and frame the canvas */
    canvas = gtk_drawing_area_new();
    gtk_signal_connect(GTK_OBJECT(canvas), "configure_event",
                       GTK_SIGNAL_FUNC(handle_canvas_resize), NULL);
    gtk_signal_connect(GTK_OBJECT(canvas), "expose_event", GTK_SIGNAL_FUNC(handle_canvas_expose),
                       NULL);
    gtk_signal_connect(GTK_OBJECT(canvas), "button_press_event",
                       GTK_SIGNAL_FUNC(handle_mouse_press), NULL);
    gtk_signal_connect(GTK_OBJECT(canvas), "button_release_event",
                       GTK_SIGNAL_FUNC(handle_mouse_release), NULL);
    gtk_widget_set_events(canvas, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
    gtk_box_pack_start(GTK_BOX(hbox), canvas, TRUE, TRUE, 0);

    /* Create a vertical scrollbar for the canvas. A number of these settings won't get properly
     * set until handle_canvas_resize is called, so use dummy values for now.
     */
    scroll_params = gtk_adjustment_new(INIT_Y_CENTER, 0, WORLD_SIZE-1, 1, 1, 0);
    vscrollbar = gtk_vscrollbar_new(GTK_ADJUSTMENT(scroll_params));
    gtk_box_pack_start(GTK_BOX(hbox), vscrollbar, FALSE, FALSE, 0);

    /* add the hbox to the containing box */
    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

    /* Create and add a horizontal scrollbar for the canvas. */
    scroll_params = gtk_adjustment_new(INIT_X_CENTER, 0, WORLD_SIZE-1, 1, 1, 0);
    hscrollbar = gtk_hscrollbar_new(GTK_ADJUSTMENT(scroll_params));
    gtk_box_pack_start(GTK_BOX(vbox), hscrollbar, FALSE, FALSE, 0);

    /* Put the canvas+scrollbars into the box */
    gtk_box_pack_start(GTK_BOX(containing_box), vbox, TRUE, TRUE, 0);
}

/* Misc. GUI functions */

/* Adjust horizontal and vertical scrollbars for the current viewport start and size.
 */
void adjust_scrollbars(void)
{
    GtkObject*  scroll_params;

    /* Necessary to say WORLD_SIZE instead of WORLD_SIZE-1 to get full range. Why? */

    scroll_params = gtk_adjustment_new(viewport_xstart, 0, WORLD_SIZE, viewport_width/8,
                                       viewport_width, viewport_width);
    gtk_range_set_adjustment(GTK_RANGE(hscrollbar), GTK_ADJUSTMENT(scroll_params));
    gtk_signal_connect(GTK_OBJECT(scroll_params), "value_changed",
                       GTK_SIGNAL_FUNC(handle_hscrollbar_change), NULL);

    scroll_params = gtk_adjustment_new(viewport_ystart, 0, WORLD_SIZE, viewport_height/8,
                                       viewport_height, viewport_height);
    gtk_range_set_adjustment(GTK_RANGE(vscrollbar), GTK_ADJUSTMENT(scroll_params));
    gtk_signal_connect(GTK_OBJECT(scroll_params), "value_changed",
                       GTK_SIGNAL_FUNC(handle_vscrollbar_change), NULL);
    gtk_widget_queue_draw(hscrollbar);
    gtk_widget_queue_draw(vscrollbar);
}

/* Adjust horizontal and vertical scrollbars for the current viewport start */
void adjust_scrollbar_values(void)
{
    GtkAdjustment*  sb_state;

    sb_state = gtk_range_get_adjustment(GTK_RANGE(hscrollbar));
    gtk_adjustment_set_value(sb_state, viewport_xstart);
    sb_state = gtk_range_get_adjustment(GTK_RANGE(vscrollbar));
    gtk_adjustment_set_value(sb_state, viewport_ystart);
}

/* Display an error/warning dialog with the given message and an "Okay" button. The message may be
 * specified in printf-style fashion, with format string and arguments.
 */
void error_dialog(const char* format, ...)
{
    va_list     args;
    GtkWidget*  dialog;
    GtkWidget*  message;
    GtkWidget*  hbox;
    GtkWidget*  error_pixmap;
    GtkWidget*  ok_button;
    char*       str;

    va_start(args, format);
    str = vdsprintf(format, args);
    va_end(args);

    /* Set up the dialog window */
    dialog = gtk_dialog_new();
    gtk_window_set_title(GTK_WINDOW(dialog), TITLE " Error");
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_container_set_border_width(GTK_CONTAINER(dialog), 5);
    gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 3);

    hbox = gtk_hbox_new(FALSE, 0);

    /* Set up the message with an error icon next to it */
    error_pixmap = load_pixmap_from_rgba_array(error_icon, ICON_WIDTH, ICON_HEIGHT,
                                               &(window->style->bg[GTK_STATE_NORMAL]));
    gtk_box_pack_start(GTK_BOX(hbox), error_pixmap, FALSE, FALSE, 5);
    message = gtk_label_new(str);
    gtk_label_set_justify(GTK_LABEL(message), GTK_JUSTIFY_LEFT);
    gtk_box_pack_start(GTK_BOX(hbox), message, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);

    /* Set up the "okay" button */
    ok_button = gtk_button_new_with_label("Okay");
    gtk_signal_connect_object(GTK_OBJECT(ok_button), "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), ok_button, FALSE, FALSE, 0);
    GTK_WIDGET_SET_FLAGS(ok_button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(ok_button);

    free(str);
    /* Show the dialog */
    gtk_widget_show_all(dialog);
}

/* Redraw the entire Life pixmap, then trigger an expose event on the canvas. Also set
 * skipped_frames back to 0.
 */
void full_canvas_redraw(void)
{
    skipped_frames = 0;
    draw_life_pixmap();
    gtk_widget_queue_draw(canvas);
}

/* Handle a mouse button click on the Life canvas.
 */
void handle_mouse_press(GtkWidget *widget, GdkEventButton *event)
{
    int32  grid_block_size;
    int32  x, y;
    int32  lx, ly;

    x = ROUND(event->x);
    y = ROUND(event->y);
    grid_block_size = (DRAW_GRID(zoom) ? zoom+1 : zoom);
    lx = viewport_xstart + x / grid_block_size;
    ly = viewport_ystart + y / grid_block_size;

    if (event->button == 1)
        /* Start drawing */
        user_draw(lx, ly);
    else if (event->button == 3) {
        /* Recenter at this point */
        set_viewport_position(lx, ly, TRUE);
        adjust_scrollbar_values();
        full_canvas_redraw();
    }
}

/* Handle a mouse button release on the Life canvas.
 */
void handle_mouse_release(GtkWidget *widget, GdkEventButton *event)
{
    if (event->button == 1)
        dragging_mouse = FALSE;
}

/* Handle mouse drag on the Life canvas. This is not actually an event handler, but is called
 * manually from the main loop if a button 1 press event was registered and a button 1 release has
 * not yet been registered.
 */
void handle_mouse_drag(void)
{
    int32  grid_block_size;
    int32  x, y;
    int32  lx, ly;

    gtk_widget_get_pointer(canvas, &x, &y);
    if (x < 0 || y < 0 || x >= canvas_width || y >= canvas_height) {
        dragging_mouse = FALSE;
        return;
    }
    grid_block_size = (DRAW_GRID(zoom) ? zoom+1 : zoom);
    lx = viewport_xstart + x / grid_block_size;
    ly = viewport_ystart + y / grid_block_size;
    if (lx != last_x || ly != last_y)
        user_draw(lx, ly);
}

/* Called when all or some of the Life display needs to be drawn onscreen. Redraw the exposed
 * rectangle from the offscreen pixmap.
 */
void handle_canvas_expose(GtkWidget *widget, GdkEventExpose *event)
{
    gdk_draw_indexed_image(canvas->window, canvas->style->fg_gc[GTK_WIDGET_STATE(canvas)],
                           event->area.x, event->area.y, event->area.width, event->area.height,
                           GDK_RGB_DITHER_NONE,
                           life_pixmap + event->area.y*canvas_width + event->area.x,
                           canvas_width, life_pixmap_colormap);
}

/* Called when the window dimensions are changed (including at startup): Set up various globals
 * dependent on window size, and recreate and redraw the canvas pixmap. Since an expose event will
 * already have been generated, we don't generate one here.
 */
void handle_canvas_resize(void)
{
    int32       xcenter, ycenter;
    int32       i;

    /* Get canvas dimensions (in pixels) */
    canvas_width  = canvas->allocation.width;
    canvas_height = canvas->allocation.height;

    /* Determine viewport dimensions (in cells) at all zoom levels for the new window size */
    for (i=MIN_ZOOM; i <= MAX_ZOOM; i *= 2) {
        if (DRAW_GRID(i)) {
            viewport_widths[i]  = (canvas_width  - 1) / (i + 1);
            viewport_heights[i] = (canvas_height - 1) / (i + 1);
        }
        else {
            viewport_widths[i]  = canvas_width  / i;
            viewport_heights[i] = canvas_height / i;
        }
    }

    /* Determine the original viewport center point. If the program has just started up, pick the
     * middle of the world.
     */
    if (!life_pixmap) {   /* program just started */
        xcenter = INIT_X_CENTER;
        ycenter = INIT_Y_CENTER;
    } else {
        xcenter = viewport_xstart + viewport_width/2;
        ycenter = viewport_ystart + viewport_height/2;
    }

    /* Record new viewport dimensions at the current zoom level */
    viewport_width = viewport_widths[zoom];
    viewport_height = viewport_heights[zoom];

    /* Set new viewport start to center around the same point as before */
    set_viewport_position(xcenter, ycenter, TRUE);

    /* Adjust scrollbars for the new viewport size and position */
    adjust_scrollbars();

    /* Create and draw a new pixmap for the new viewport */
    if (canvas_width * canvas_height > life_pixmap_alloc_len) {
        life_pixmap_alloc_len = canvas_width * canvas_height;
        life_pixmap = safe_realloc(life_pixmap, life_pixmap_alloc_len);
    }
    draw_life_pixmap();
}

/* Called when the horizontal scrollbar is adjusted (by the user or by our program). Adjust the
 * viewport start, and do a full redraw.
 */
void handle_hscrollbar_change(void)
{
    GtkAdjustment*  sb_state;

    sb_state = gtk_range_get_adjustment(GTK_RANGE(hscrollbar));
    set_viewport_position(ROUND(sb_state->value), viewport_ystart, FALSE);
    full_canvas_redraw();
}

/* Called when the vertical scrollbar is adjusted (by the user or by our program). Adjust the
 * viewport start, and do a full redraw.
 */
void handle_vscrollbar_change(void)
{
    GtkAdjustment*  sb_state;

    sb_state = gtk_range_get_adjustment(GTK_RANGE(vscrollbar));
    set_viewport_position(viewport_xstart, ROUND(sb_state->value), FALSE);
    full_canvas_redraw();
}

/* Called when the user selects a pattern file from the sidebar */
void handle_sidebar_select(GtkCList *clist, gint row, gint column, GdkEventButton *event,
                           gpointer user_data)
{
    char*  filename;
    char*  path;

    if (row < 0)
        return;
    if (gtk_clist_get_text(GTK_CLIST(clist), row, column, &filename)) {
        path = dsprintf("%s%s.lif", patterns_dir, filename);
        attempt_load_pattern(path);
        free(path);
    }
}

/* Called when the speed slider has been adjusted (by the user or by our program). Set the new
 * speed, update the speed label, and reset the FPS measurement if the pattern is running.
 */
void handle_speed_slider_change(void)
{
    GtkAdjustment*  slider_state;
    char*           speed_str;

    slider_state = gtk_range_get_adjustment(GTK_RANGE(speed_slider));
    speed = ROUND(slider_state->value);
    speed_str = dsprintf("%d", speed);
    gtk_label_set_text(GTK_LABEL(speed_label), speed_str);
    free(speed_str);

    /* Reset the running FPS measurement if necessary */
    if (pattern_running) {
        start_tick = tick;
        start_time = get_time_milliseconds();
    }
}

/* Called when the zoom level has changed by some means. Update various globals dependent on zoom
 * level, then do a full redraw.
 */
void handle_zoom_change(void)
{
    int32  xcenter, ycenter;

    /* Determine the original viewport center point */
    xcenter = viewport_xstart + viewport_width/2;
    ycenter = viewport_ystart + viewport_height/2;

    /* Record new viewport dimensions at the current zoom level */
    viewport_width = viewport_widths[zoom];
    viewport_height = viewport_heights[zoom];

    /* Set new viewport start to center around the same point as before */
    set_viewport_position(xcenter, ycenter, TRUE);

    /* Adjust scrollbars for the new viewport size and position */
    adjust_scrollbars();

    /* Draw the pixmap and generate a canvas expose */
    full_canvas_redraw();
}

/* Put the current filename (taken from recent_files[0]) into the main window title */
void put_filename_in_window_title(void)
{
    char*  title;

    title = dsprintf("%s: %s", TITLE, recent_files[0].filename);
    gtk_window_set_title(GTK_WINDOW(window), title);
    free(title);
}

/* Adjust horizontal widget size for the speed label, based on max speed */
void set_speed_label_size(int32 max_speed)
{
    char*  speed_str;

    speed_str = dsprintf("%d", max_speed);
    gtk_widget_set_usize(GTK_WIDGET(speed_label),
                         gdk_string_measure(window->style->font, speed_str), 0);
    free(speed_str);
}

/* Set the global variables for viewport start and end, and let the backend know about the new
 * viewport. If center_around is true, attempt to center around the given point, otherwise try to
 * set the given point at the upper left corner. If the viewport would go off the edge of the
 * world, readjust it.
 */
void set_viewport_position(int32 x, int32 y, boolean center_around)
{
    /* Set viewport start */
    if (center_around) {
        viewport_xstart = x - viewport_width/2;
        viewport_ystart = y - viewport_height/2;

    } else {
        viewport_xstart = x;
        viewport_ystart = y;
    }

    /* Readjust viewport start if the current viewport goes off the edge of the world */
    if (viewport_xstart < 0)
        viewport_xstart = 0;
    else if (viewport_xstart + viewport_width > WORLD_SIZE)
        viewport_xstart = WORLD_SIZE - viewport_width;
    if (viewport_ystart < 0)
        viewport_ystart = 0;
    else if (viewport_ystart + viewport_height > WORLD_SIZE)
        viewport_ystart = WORLD_SIZE - viewport_height;

    /* Set viewport end */
    viewport_xend = viewport_xstart + viewport_width  - 1;
    viewport_yend = viewport_ystart + viewport_height - 1;

    /* Let the backend know about the new viewport */
    set_viewport(viewport_xstart, viewport_ystart, viewport_xend, viewport_yend);
}

/* Trigger an expose event on the portion of the Life canvas which has changed over the last tick,
 * setting skipped_frames back to 0.
 */
void trigger_canvas_update(void)
{
    skipped_frames = 0;
    if (update_xstart <= update_xend && update_ystart <= update_yend)
        gtk_widget_queue_draw_area(canvas, update_xstart, update_ystart,
                                   update_xend - update_xstart + 1,
                                   update_yend - update_ystart + 1);
}

/* Update the onscreen population label to match the current population. */
void update_population_label(void)
{
    char  str[20];

    sprintf(str, "%u", population);
    gtk_label_set_text(GTK_LABEL(population_label), str);
}

/* Regenerate the list of patterns for the sidebar. */
void update_pattern_list(void)
{
    DIR*            dir;
    struct dirent*  file;
    char*           filename;
    int32           num_patterns = 0;
    int32           ext_len;
    int32           len;

    gtk_clist_freeze(GTK_CLIST(pattern_list));
    gtk_clist_clear(GTK_CLIST(pattern_list));

    ext_len = strlen(DEFAULT_EXTENSION);
    dir = opendir(patterns_dir);
    if (dir) {
        while ((file = readdir(dir)) != NULL) {
            filename = safe_strdup(file->d_name);
            len = strlen(filename);
            if (len > ext_len + 1 && STR_EQUAL(filename + len - ext_len, DEFAULT_EXTENSION)) {
                filename[len - ext_len - 1] = '\0';
                gtk_clist_append(GTK_CLIST(pattern_list), &filename);
                num_patterns++;
            }
            free(filename);
        }
        closedir(dir);
    }

    gtk_clist_sort(GTK_CLIST(pattern_list));
    gtk_widget_set_usize(pattern_list,
                         gtk_clist_optimal_column_width(GTK_CLIST(pattern_list), 0) + 40, 0);

    gtk_clist_thaw(GTK_CLIST(pattern_list));
}

/* Update the onscreen tick label to match the current tick. */
void update_tick_label(void)
{
    char  str[20];

    sprintf(str, "%u", tick);
    gtk_label_set_text(GTK_LABEL(tick_label), str);
}

/* Create an integral, snap-to-ticks spinbutton with the given initial, minimum and maximum values.
 */
GtkWidget* create_spinbox(int32 init_value, int32 min, int32 max)
{
    GtkWidget*  spinbox;
    GtkObject*  adjustment;

    adjustment = gtk_adjustment_new(init_value, min, max, 1, 1, 0);
    spinbox = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 1, 0);
    gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(spinbox), TRUE);
    gtk_spin_button_set_snap_to_ticks(GTK_SPIN_BUTTON(spinbox), TRUE);

    return spinbox;
}

/* Functions for drawing the offscreen pixmap */

/* Fully redraw the Life pixmap: backdrop, grid, and live cells.
 */
void draw_life_pixmap(void)
{
    cage_type*  c;
    boolean     processed_sleepers = FALSE;
    int32       hor_line_width, vert_line_height;
    int32       x, y;

    /* Draw the backdrop */
    memset(life_pixmap, BG_COLOR_INDEX, canvas_width*canvas_height);

    /* Draw the grid, if necessary */
    if (DRAW_GRID(zoom)) {
        hor_line_width   = viewport_width  * (zoom+1) + 1;
        vert_line_height = viewport_height * (zoom+1) + 1;
        /* Draw the horizontal lines */
        for (y=0; y < canvas_height; y += (zoom+1))
            memset(life_pixmap + y*canvas_width, GRID_COLOR_INDEX, hor_line_width);
        /* Draw the vertical lines */
        for (x=0; x < canvas_width; x += (zoom+1)) {
            for (y=0; y < vert_line_height; y++)
                life_pixmap[y*canvas_width + x] = GRID_COLOR_INDEX;
        }
    }

    /* Draw the live cells */
    for (c = active_cages; ; c = c->next) {
        if (!c) {
            if (!processed_sleepers) {
                c = sleeping_cages;
                processed_sleepers = TRUE;
            }
            if (!c)
                break;
        }

        if (IS_ONSCREEN(c)) {
            x = c->x * CAGE_SIZE;
            y = c->y * CAGE_SIZE;
            if (c->bnw[parity])
                draw_life_block(c->bnw[parity], x, y, TRUE);
            if (c->bne[parity])
                draw_life_block(c->bne[parity], x + BLOCK_SIZE, y, TRUE);
            if (c->bsw[parity])
                draw_life_block(c->bsw[parity], x, y + BLOCK_SIZE, TRUE);
            if (c->bse[parity])
                draw_life_block(c->bse[parity], x + BLOCK_SIZE, y + BLOCK_SIZE, TRUE);
        }
    }
}

/* Draw the given 4x4 cell block onto the Life pixmap, at the given starting coordinates.
 * If full_update is TRUE, draw only live cells (because the screen has been cleared beforehand),
 * otherwise draw all cells. If full_update is FALSE, update the current clipping rectangle
 * (update_xstart, etc.) to include this block.
 */
void draw_life_block(uint16 block, int32 xstart, int32 ystart, boolean full_update)
{
    int32     xend, yend;
    int32     minx, miny, maxx, maxy;
    int32     real_minx, real_miny, real_maxx, real_maxy;
    int32     realx, realy;
    int32     grid_block_size;
    int32     start_bit;
    int32     bit;
    int32     x, y, i;
    uint8     color;
    uint8*    spos;
    uint8*    pos;

    grid_block_size = (DRAW_GRID(zoom) ? zoom+1 : zoom);

    /* Figure out the screen offset, in blocks */
    xstart -= viewport_xstart;
    ystart -= viewport_ystart;
    xend = xstart + BLOCK_SIZE - 1;
    yend = ystart + BLOCK_SIZE - 1;
    minx = MAX(xstart, 0);
    miny = MAX(ystart, 0);
    maxx = MIN(xend, viewport_width-1);
    maxy = MIN(yend, viewport_height-1);
    if (minx > maxx || miny > maxy)
        return;

    /* Update the clipping rectangle, if necessary */
    if (!full_update) {
        real_minx = minx * grid_block_size;
        real_miny = miny * grid_block_size;
        real_maxx = maxx * grid_block_size + (zoom - 1);
        real_maxy = maxy * grid_block_size + (zoom - 1);
        if (DRAW_GRID(zoom)) {
            real_minx++;
            real_miny++;
            real_maxx++;
            real_maxy++;
        }
        if (real_minx < update_xstart)
            update_xstart = real_minx;
        if (real_miny < update_ystart)
            update_ystart = real_miny;
        if (real_maxx > update_xend)
            update_xend = real_maxx;
        if (real_maxy > update_yend)
            update_yend = real_maxy;
    }

    start_bit = (miny - ystart) * BLOCK_SIZE + (minx - xstart);
    /* Draw the cells */
    if (zoom > 1) {
        for (y=miny; y <= maxy; y++, start_bit += BLOCK_SIZE) {
            for (x=minx, bit=start_bit; x <= maxx; x++, bit++) {
                color = ((block & (1 << bit)) ? CELL_COLOR_INDEX : BG_COLOR_INDEX);
                if (full_update && color == BG_COLOR_INDEX)
                    continue;
                realx = x * grid_block_size;
                realy = y * grid_block_size;
                if (DRAW_GRID(zoom)) {
                    realx++;
                    realy++;
                }
                for (i=0, pos=life_pixmap + realy*canvas_width + realx;
                     i < zoom;
                     i++, pos += canvas_width)
                    memset(pos, color, zoom);
            }
        }
    } else {
        spos = life_pixmap + miny * canvas_width + minx;
        for (y=miny; y <= maxy; y++, spos += canvas_width, start_bit += BLOCK_SIZE) {
            for (x=minx, bit=start_bit, pos=spos; x <= maxx; x++, pos++, bit++) {
                color = ((block & (1 << bit)) ? CELL_COLOR_INDEX : BG_COLOR_INDEX);
                *pos = color;
            }
        }
    }
}

/* Set/unset the cell at the given absolute (logical) coordinates, then update the Life state and
 * the screen. If the coordinate are outside of the viewport, adjust them. If the user just clicked
 * the mouse, check the color of the cell to determine whether to set or erase. If they've been
 * dragging the mouse, continue either setting or erasing.
 */
void user_draw(int32 x, int32 y)
{
    static  uint8  color;
    uint8*  pos;
    uint8   cur_color;
    int32   grid_block_size;
    int32   screenx, screeny;
    int32   i;

    if (x < viewport_xstart)
        x = viewport_xstart;
    else if (x >= viewport_xstart + viewport_width)
        x = viewport_xstart + viewport_width - 1;
    if (y < viewport_ystart)
        y = viewport_ystart;
    else if (y >= viewport_ystart + viewport_height)
        y = viewport_ystart + viewport_height - 1;
    last_x = x;
    last_y = y;

    grid_block_size = (DRAW_GRID(zoom) ? zoom+1 : zoom);
    screenx = (x - viewport_xstart) * grid_block_size;
    screeny = (y - viewport_ystart) * grid_block_size;
    if (DRAW_GRID(zoom)) {
        screenx++;
        screeny++;
    }

    cur_color = life_pixmap[screeny*canvas_width + screenx];
    if (dragging_mouse) {
        if (cur_color == color)
            return;
    } else
        color = ((cur_color == CELL_COLOR_INDEX) ? BG_COLOR_INDEX : CELL_COLOR_INDEX);

    draw_cell(x, y, (color == BG_COLOR_INDEX ? DRAW_UNSET : DRAW_SET));
    if (zoom == 1)
        life_pixmap[screeny*canvas_width + screenx] = color;
    else {
        for (i=0, pos = life_pixmap + screeny*canvas_width + screenx;
             i < zoom;
             i++, pos += canvas_width)
            memset(pos, color, zoom);
    }

    gtk_widget_queue_draw_area(canvas, screenx, screeny, zoom, zoom);
    dragging_mouse = TRUE;

    update_population_label();
}

/* Bound Functions */

void file_new(void)
{
    if (pattern_running)
        start_stop();
    clear_world();
    position_center();   /* will trigger a canvas redraw */
    update_tick_label();
    update_population_label();
    sidebar_unselect();
    free(pattern_path);
    pattern_path = NULL;
    gtk_window_set_title(GTK_WINDOW(window), TITLE);
}

void file_open(void)
{
    GtkWidget*  file_selector;

    file_selector = gtk_file_selection_new("Load Pattern");
    gtk_window_set_position(GTK_WINDOW(file_selector), GTK_WIN_POS_CENTER);
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_selector), current_dir);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->ok_button), "clicked",
                       GTK_SIGNAL_FUNC(file_open_ok), file_selector);
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->cancel_button),
                              "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(file_selector));
    gtk_widget_show(file_selector);
}

void file_reopen(void)
{
    if (pattern_path)
        attempt_load_pattern(pattern_path);
    else
        error_dialog("Reopen failed: no current file");
}

void file_save(void)
{
    if (pattern_path)
        attempt_save_pattern(pattern_path);
    else
        file_save_as();
}

void file_save_as(void)
{
    GtkWidget*  file_selector;
    char*       home_dir;

    file_selector = gtk_file_selection_new("Save Pattern");
    gtk_window_set_position(GTK_WINDOW(file_selector), GTK_WIN_POS_CENTER);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->ok_button), "clicked",
                       GTK_SIGNAL_FUNC(file_save_as_ok), file_selector);
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->cancel_button),
                              "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(file_selector));
    if (pattern_path)
        gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_selector), pattern_path);
    else {
        home_dir = getenv("HOME");
        if (home_dir) {
            home_dir = dsprintf("%s/", home_dir);
            gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_selector), home_dir);
            free(home_dir);
        }
    }
    gtk_widget_show(file_selector);
}

void file_description(void)
{
    GtkWidget*  dialog;
    GtkWidget*  textbox;
    GtkWidget*  scrollbar;
    GtkWidget*  button;
    GtkWidget*  alignment;
    GtkWidget*  hbox;
    desc_dialog_info*  info;
    int32  desc_width = 0;
    int32  desc_height;
    int32  w;
    int32  i;

    info = safe_malloc(sizeof(desc_dialog_info));

    /* Determine the width and height of the current description text */
    for (i=0; i < desc_num_lines; i++) {
        w = gdk_string_measure(window->style->font, pattern_description[i]) + 40;
        if (w > desc_width)
            desc_width = w;
    }
    if (desc_width == 0) {
        desc_width = 500;
        desc_height = 200;
    } else {
        if (desc_width < 300)
            desc_width = 300;
        desc_height = MIN(400, 20 * desc_num_lines);
    }

    /* Set up the dialog window */
    dialog = gtk_dialog_new();
    gtk_window_set_title(GTK_WINDOW(dialog), "View/Edit Pattern Description");
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_window_set_default_size(GTK_WINDOW(dialog), desc_width, desc_height);
    gtk_container_set_border_width(GTK_CONTAINER(dialog), 5);
    gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 3);
    gtk_signal_connect(GTK_OBJECT(dialog), "destroy", GTK_SIGNAL_FUNC(file_description_destroy),
                       info);

    hbox = gtk_hbox_new(FALSE, 0);

    /* Set up the textbox with vertical scrollbar */
    textbox = gtk_text_new(NULL, NULL);
    gtk_text_set_editable(GTK_TEXT(textbox), TRUE);
    gtk_text_set_word_wrap(GTK_TEXT(textbox), TRUE);
    for (i=0; i < desc_num_lines; i++) {
        gtk_text_insert(GTK_TEXT(textbox), NULL, NULL, NULL, pattern_description[i], -1);
        gtk_text_insert(GTK_TEXT(textbox), NULL, NULL, NULL, "\n", -1);
    }
    gtk_box_pack_start(GTK_BOX(hbox), textbox, TRUE, TRUE, 0);
    scrollbar = gtk_vscrollbar_new(GTK_TEXT(textbox)->vadj);
    gtk_box_pack_start(GTK_BOX(hbox), scrollbar, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);

    /* Data to be sent to signal handlers for this dialog */
    info->dialog  = dialog;
    info->textbox = textbox;

    /* Set up the buttons */
    button = gtk_button_new_with_label("Okay");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(file_description_ok), info);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(button);
    button = gtk_button_new_with_label("Apply");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(file_description_apply),
                       info);
    CENTER_ALIGN(button);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), alignment, FALSE, FALSE, 0);
    button = gtk_button_new_with_label("Cancel");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
    CENTER_ALIGN(button);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), alignment, FALSE, FALSE, 0);

    /* Show the dialog */
    gtk_widget_show_all(dialog);
}

void file_recent(GtkWidget* widget, int32* num)
{
    if (recent_files[*num].full_path) {
        if (attempt_load_pattern(recent_files[*num].full_path))
            sidebar_unselect();
    }
}

void file_quit(void)
{
    save_keybindings();
    save_recent_files_list();
    exit(0);
}

void view_zoom_in(void)
{
    if (zoom < MAX_ZOOM) {
        zoom *= 2;
        handle_zoom_change();
    }
}

void view_zoom_out(void)
{
    if (zoom > MIN_ZOOM) {
        zoom /= 2;
        handle_zoom_change();
    }
}

void view_zoom_1(void)
{
    if (zoom != 1) {
        zoom = 1;
        handle_zoom_change();
    }
}

void view_zoom_2(void)
{
    if (zoom != 2) {
        zoom = 2;
        handle_zoom_change();
    }
}

void view_zoom_4(void)
{
    if (zoom != 4) {
        zoom = 4;
        handle_zoom_change();
    }
}

void view_zoom_8(void)
{
    if (zoom != 8) {
        zoom = 8;
        handle_zoom_change();
    }
}

void view_zoom_16(void)
{
    if (zoom != 16) {
        zoom = 16;
        handle_zoom_change();
    }
}

void view_show_toolbar(void)
{
    if (GTK_CHECK_MENU_ITEM(show_toolbar_menu_item)->active == show_toolbar)
        return;

    show_toolbar = !show_toolbar;
    if (show_toolbar)
        gtk_widget_show_all(toolbar_box);
    else
        gtk_widget_hide_all(toolbar_box);
}

void view_show_sidebar(void)
{
    if (GTK_CHECK_MENU_ITEM(show_sidebar_menu_item)->active == show_sidebar)
        return;

    show_sidebar = !show_sidebar;
    if (show_sidebar)
        gtk_widget_show_all(sidebar);
    else
        gtk_widget_hide_all(sidebar);
}

void view_show_scrollbars(void)
{
    if (GTK_CHECK_MENU_ITEM(show_scrollbars_menu_item)->active == show_scrollbars)
        return;

    show_scrollbars = !show_scrollbars;
    if (show_scrollbars) {
        gtk_widget_show(vscrollbar);
        gtk_widget_show(hscrollbar);
    } else {
        gtk_widget_hide(vscrollbar);
        gtk_widget_hide(hscrollbar);
    }
}

void position_center(void)
{
    set_viewport_position(WORLD_SIZE/2, WORLD_SIZE/2, TRUE);
    adjust_scrollbar_values();
    full_canvas_redraw();
}

void position_scroll_left(void)
{
    position_scroll_generic(-1, 0, FALSE);
}

void position_scroll_right(void)
{
    position_scroll_generic(1, 0, FALSE);
}

void position_scroll_up(void)
{
    position_scroll_generic(0, -1, FALSE);
}

void position_scroll_down(void)
{
    position_scroll_generic(0, 1, FALSE);
}

void position_scroll_nw(void)
{
    position_scroll_generic(-1, -1, FALSE);
}

void position_scroll_ne(void)
{
    position_scroll_generic(1, -1, FALSE);
}

void position_scroll_sw(void)
{
    position_scroll_generic(-1, 1, FALSE);
}

void position_scroll_se(void)
{
    position_scroll_generic(1, 1, FALSE);
}

void position_page_left(void)
{
    position_scroll_generic(-1, 0, TRUE);
}

void position_page_right(void)
{
    position_scroll_generic(1, 0, TRUE);
}

void position_page_up(void)
{
    position_scroll_generic(0, -1, TRUE);
}

void position_page_down(void)
{
    position_scroll_generic(0, 1, TRUE);
}

void position_page_nw(void)
{
    position_scroll_generic(-1, -1, TRUE);
}

void position_page_ne(void)
{
    position_scroll_generic(1, -1, TRUE);
}

void position_page_sw(void)
{
    position_scroll_generic(-1, 1, TRUE);
}

void position_page_se(void)
{
    position_scroll_generic(1, 1, TRUE);
}

void edit_preferences(void)
{
    GtkWidget*  dialog;
    GtkWidget*  tabbed_frame;
    GtkWidget*  table;
    GtkWidget*  vbox;
    GtkWidget*  hbox;
    GtkWidget*  frame;
    GtkWidget*  alignment;
    GtkWidget*  button;
    GtkWidget*  label;
    prefs_dialog_info*  info;
    color_dialog_info*  color_info;
    GList*      zoom_list = NULL;
    GList*      pos;
    char*       title_text;
    char*       label_text;
    char*       default_zoom_str;
    char*       buf;
    uint32      default_value;
    int32       i;

    /* Setup the dialog window */
    dialog = gtk_dialog_new();
    info = safe_malloc(sizeof(prefs_dialog_info));
    info->dialog = dialog;
    gtk_window_set_title(GTK_WINDOW(dialog), TITLE " Preferences");
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_signal_connect(GTK_OBJECT(dialog), "destroy", GTK_SIGNAL_FUNC(edit_preferences_destroy),
                       info);

    /* Init the tabbed frame */
    tabbed_frame = gtk_notebook_new();

    /*** Setup "File" preferences ***/

    frame = gtk_frame_new("General");
    gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
    table = gtk_table_new(1, 4, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 3);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    /* Patterns directory */
    LEFT_ALIGN(gtk_label_new("Patterns directory:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);
    info->patterns_dir_entry = gtk_entry_new_with_max_length(PATH_MAX);
    gtk_widget_set_usize(info->patterns_dir_entry, 200, 0);
    gtk_entry_set_text(GTK_ENTRY(info->patterns_dir_entry), patterns_dir);
    gtk_editable_set_position(GTK_EDITABLE(info->patterns_dir_entry), 0);
    LEFT_ALIGN(info->patterns_dir_entry);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);
    button = gtk_button_new_with_label("....");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(prefs_select_patterns_dir),
                       info);
    LEFT_ALIGN(button);
    gtk_table_attach(GTK_TABLE(table), alignment, 2, 3, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);

    button = gtk_button_new_with_label(" Default ");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(prefs_default_patterns_dir),
                       info);
    LEFT_ALIGN(button);
    gtk_table_attach(GTK_TABLE(table), alignment, 3, 4, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Add "File" prefs to the tabbed frame */
    gtk_container_add(GTK_CONTAINER(frame), table);
    FRAME_ALIGN(frame);
    gtk_notebook_append_page(GTK_NOTEBOOK(tabbed_frame), alignment, gtk_label_new("File"));

    /*** Setup "View" preferences ***/

    vbox = gtk_vbox_new(FALSE, 0);

    /** "General" View Prefs **/

    frame = gtk_frame_new("General");
    gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
    table = gtk_table_new(5, 2, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 3);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    /* Initial window size */
    LEFT_ALIGN(gtk_label_new("Initial window size:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);
    hbox = gtk_hbox_new(FALSE, 0);
    info->init_window_width_spinbox = create_spinbox(init_window_width, 0, 10000);
    gtk_box_pack_start(GTK_BOX(hbox), info->init_window_width_spinbox, FALSE, FALSE, 0);
    label = gtk_label_new("X");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
    info->init_window_height_spinbox = create_spinbox(init_window_height, 0, 10000);
    gtk_box_pack_start(GTK_BOX(hbox), info->init_window_height_spinbox, FALSE, FALSE, 0);
    LEFT_ALIGN(hbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Default zoom level */
    LEFT_ALIGN(gtk_label_new("Default zoom level:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 1, 2, TABLE_OPTS, TABLE_OPTS, 0, 0);
    info->default_zoom_combo = gtk_combo_new();
    gtk_widget_set_usize(info->default_zoom_combo, 75, 0);
    default_zoom_str = "1:1";
    for (i=MIN_ZOOM; i <= MAX_ZOOM; i*=2) {
        buf = dsprintf("%d:1", i);
        zoom_list = g_list_append(zoom_list, buf);
        if (i == default_zoom)
            default_zoom_str = buf;
    }
    gtk_combo_set_popdown_strings(GTK_COMBO(info->default_zoom_combo), zoom_list);
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(info->default_zoom_combo)->entry), default_zoom_str);
    for (pos=zoom_list; pos; pos=g_list_next(pos))
        free(pos->data);
    g_list_free(zoom_list);
    LEFT_ALIGN(info->default_zoom_combo);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 1, 2, TABLE_OPTS, TABLE_OPTS, 0, 0);

    LEFT_ALIGN(gtk_label_new("Show at startup:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 2, 3, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Default show toolbar */
    hbox = gtk_hbox_new(FALSE, 0);
    info->default_show_toolbar_toggle = gtk_check_button_new();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(info->default_show_toolbar_toggle),
                                 default_show_toolbar);
    gtk_box_pack_start(GTK_BOX(hbox), info->default_show_toolbar_toggle, FALSE, FALSE, 0);
    label = gtk_label_new("Toolbar");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);
    LEFT_ALIGN(hbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 2, 3, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Default show sidebar */
    hbox = gtk_hbox_new(FALSE, 0);
    info->default_show_sidebar_toggle = gtk_check_button_new();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(info->default_show_sidebar_toggle),
                                 default_show_sidebar);
    gtk_box_pack_start(GTK_BOX(hbox), info->default_show_sidebar_toggle, FALSE, FALSE, 0);
    label = gtk_label_new("Sidebar");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);
    LEFT_ALIGN(hbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 3, 4, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Default show scrollbars */
    hbox = gtk_hbox_new(FALSE, 0);
    info->default_show_scrollbars_toggle = gtk_check_button_new();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(info->default_show_scrollbars_toggle),
                                 default_show_scrollbars);
    gtk_box_pack_start(GTK_BOX(hbox), info->default_show_scrollbars_toggle, FALSE, FALSE, 0);
    label = gtk_label_new("Scrollbars");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);
    LEFT_ALIGN(hbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 4, 5, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Add "General" view prefs */
    gtk_container_add(GTK_CONTAINER(frame), table);
    FRAME_ALIGN(frame);
    gtk_box_pack_start(GTK_BOX(vbox), alignment, FALSE, FALSE, 3);

    /** "Colors" View Prefs **/

    frame = gtk_frame_new("Colors");
    gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
    table = gtk_table_new(3, 4, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 3);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    /* Loop through colors */
    for (i=0; i < NUM_COLORS; i++) {
        if (i == BG_COLOR_INDEX) {
            title_text    = "Choose Background Color";
            label_text    = "Background:";
            default_value = DEFAULT_BG_COLOR;
        } else if (i == CELL_COLOR_INDEX) {
            title_text  = "Choose Cell Color";
            label_text  = "Live cells:";
            default_value = DEFAULT_CELL_COLOR;
        } else {
            title_text  = "Choose Grid Color";
            label_text  = "Grid:";
            default_value = DEFAULT_GRID_COLOR;
        }
        color_info = info->color_dialog_infos[i] = safe_malloc(sizeof(color_dialog_info));

        LEFT_ALIGN(gtk_label_new(label_text));
        gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, i, i+1, TABLE_OPTS, TABLE_OPTS, 0, 0);
        info->color_entries[i] = gtk_entry_new();
        gtk_widget_set_usize(info->color_entries[i], 75, 0);
        buf = dsprintf("#%06X", colors[i]);
        gtk_entry_set_text(GTK_ENTRY(info->color_entries[i]), buf);
        free(buf);
        LEFT_ALIGN(info->color_entries[i]);
        gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, i, i+1, TABLE_OPTS, TABLE_OPTS, 0, 0);
        color_info->dialog_title  = title_text;
        color_info->prefs_entry   = info->color_entries[i];
        color_info->cur_value     = colors[i];
        color_info->default_value = default_value;
        button = gtk_button_new_with_label(" Select ");
        gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(prefs_select_color),
                           color_info);
        LEFT_ALIGN(button);
        gtk_table_attach(GTK_TABLE(table), alignment, 2, 3, i, i+1, TABLE_OPTS, TABLE_OPTS, 0, 0);

        button = gtk_button_new_with_label(" Default ");
        gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(prefs_default_color),
                           color_info);
        LEFT_ALIGN(button);
        gtk_table_attach(GTK_TABLE(table), alignment, 3, 4, i, i+1, TABLE_OPTS, TABLE_OPTS, 0, 0);
    }

    /* Add "Colors" view prefs */
    gtk_container_add(GTK_CONTAINER(frame), table);
    FRAME_ALIGN(frame);
    gtk_box_pack_start(GTK_BOX(vbox), alignment, FALSE, FALSE, 3);

    /* Add "View" prefs to the tabbed frame */
    gtk_notebook_append_page(GTK_NOTEBOOK(tabbed_frame), vbox, gtk_label_new("View"));

    /*** Setup "Run" preferences ***/

    frame = gtk_frame_new("Speed");
    gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
    table = gtk_table_new(4, 2, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 3);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    /* Default speed */
    LEFT_ALIGN(gtk_label_new("Default speed (Gen/s):"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);
    info->default_speed_spinbox = create_spinbox(default_speed, 1, 10000);
    LEFT_ALIGN(info->default_speed_spinbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Speed slider range */
    LEFT_ALIGN(gtk_label_new("Speed slider range:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 1, 2, TABLE_OPTS, TABLE_OPTS, 0, 0);
    hbox = gtk_hbox_new(FALSE, 0);
    info->speed_slider_min_spinbox = create_spinbox(speed_slider_min, 1, speed);
    gtk_box_pack_start(GTK_BOX(hbox), info->speed_slider_min_spinbox, FALSE, FALSE, 0);
    label = gtk_label_new("-");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
    info->speed_slider_max_spinbox = create_spinbox(speed_slider_max, speed, 10000);
    gtk_box_pack_start(GTK_BOX(hbox), info->speed_slider_max_spinbox, FALSE, FALSE, 3);
    LEFT_ALIGN(hbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 1, 2, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Speed increment */
    LEFT_ALIGN(gtk_label_new("Speed increment:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 2, 3, TABLE_OPTS, TABLE_OPTS, 0, 0);
    info->speed_increment_spinbox = create_spinbox(speed_increment, 1, 10000);
    LEFT_ALIGN(info->speed_increment_spinbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 2, 3, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Default skip frames */
    hbox = gtk_hbox_new(FALSE, 0);
    info->default_skip_toggle = gtk_check_button_new();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(info->default_skip_toggle),
                                 default_skip_frames);
    gtk_box_pack_start(GTK_BOX(hbox), info->default_skip_toggle, FALSE, FALSE, 0);
    label = gtk_label_new("Skip frames by default");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);
    LEFT_ALIGN(hbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 2, 3, 4, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Add "Run" prefs to the tabbed frame */
    gtk_container_add(GTK_CONTAINER(frame), table);
    FRAME_ALIGN(frame);
    gtk_notebook_append_page(GTK_NOTEBOOK(tabbed_frame), alignment, gtk_label_new("Run"));

    /*** Setup "Help Browser" preferences ***/
    frame = gtk_frame_new("General");
    gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
    table = gtk_table_new(1, 3, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 3);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    /* Browser command-line */
    LEFT_ALIGN(gtk_label_new("Command line:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);
    info->help_browser_entry = gtk_entry_new();
    gtk_widget_set_usize(info->help_browser_entry, 250, 0);
    gtk_entry_set_text(GTK_ENTRY(info->help_browser_entry), help_browser);
    gtk_editable_set_position(GTK_EDITABLE(info->help_browser_entry), 0);
    LEFT_ALIGN(info->help_browser_entry);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);

    button = gtk_button_new_with_label(" Default ");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(prefs_default_help_browser),
                       info);
    LEFT_ALIGN(button);
    gtk_table_attach(GTK_TABLE(table), alignment, 2, 3, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Add "Help Browser" prefs to the tabbed frame */
    gtk_container_add(GTK_CONTAINER(frame), table);
    FRAME_ALIGN(frame);
    gtk_notebook_append_page(GTK_NOTEBOOK(tabbed_frame), alignment, gtk_label_new("Help Browser"));

    /* Add tabbed frame to dialog vbox */
    gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 10);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), tabbed_frame, FALSE, FALSE, 5);

    /* Setup the buttons */
    button  = gtk_button_new_with_label("Okay");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(edit_preferences_ok), info);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(button);
    button  = gtk_button_new_with_label("Apply");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(edit_preferences_apply),
                       info);
    CENTER_ALIGN(button);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), alignment, FALSE, FALSE, 0);
    button = gtk_button_new_with_label("Cancel");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
    CENTER_ALIGN(button);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), alignment, FALSE, FALSE, 0);

    /* Show the dialog */
    gtk_widget_show_all(dialog);
}

void run_start_stop(void)
{
    if (pattern_running && skipped_frames) {
        trigger_canvas_update();
        update_tick_label();
        update_population_label();
    }
    start_stop();
}

void run_step(void)
{
    if (!pattern_running)
        tick_and_update();
}

void run_faster(void)
{
    set_speed(speed + speed_increment);
}

void run_slower(void)
{
    set_speed(speed - speed_increment);
}

void run_speed(void)
{
    GtkWidget*  dialog;
    GtkWidget*  speed_spinbox;
    GtkWidget*  skip_toggle;
    GtkWidget*  button;
    GtkWidget*  label;
    GtkWidget*  alignment;
    GtkWidget*  hbox;
    speed_dialog_info*  info;

    info = safe_malloc(sizeof(speed_dialog_info));

    /* Setup the dialog window */
    dialog = gtk_dialog_new();
    gtk_window_set_title(GTK_WINDOW(dialog), "Set Speed");
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_signal_connect(GTK_OBJECT(dialog), "destroy", GTK_SIGNAL_FUNC(run_speed_destroy), info);

    /* Setup the Gen/s speed spinbox */
    hbox = gtk_hbox_new(FALSE, 0);
    label = gtk_label_new("Speed in generations/sec:");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
    speed_spinbox = create_spinbox(speed, 1, 10000);
    gtk_box_pack_start(GTK_BOX(hbox), speed_spinbox, FALSE, FALSE, 5);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hbox, FALSE, FALSE, 5);

    /* Setup the "skip frames" toggle */
    hbox = gtk_hbox_new(FALSE, 0);
    skip_toggle = gtk_check_button_new();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(skip_toggle), skip_frames);
    gtk_box_pack_start(GTK_BOX(hbox), skip_toggle, FALSE, FALSE, 3);
    label = gtk_label_new("Skip frames to achieve speed");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hbox, FALSE, FALSE, 5);

    /* Data to be sent to "okay" button signal handler */
    info->dialog        = dialog;
    info->speed_spinbox = speed_spinbox;
    info->skip_toggle   = skip_toggle;

    /* Setup the buttons */
    button  = gtk_button_new_with_label("Okay");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(run_speed_ok), info);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(button);
    button = gtk_button_new_with_label("Apply");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(run_speed_apply), info);
    CENTER_ALIGN(button);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), alignment, FALSE, FALSE, 0);
    button = gtk_button_new_with_label("Cancel");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
    CENTER_ALIGN(button);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), alignment, FALSE, FALSE, 0);

    /* Show the dialog */
    gtk_widget_show_all(dialog);
}

void help_help(void)
{
    char*  url = "file://" DOCDIR "/index.html";
    char*  command;
    char*  old_command;
    char*  p;

    command = dsprintf("%s &", help_browser);
    while ((p = strstr(command, HELP_BROWSER_PARAM)) != NULL) {
        *p = '\0';
        old_command = command;
        command = dsprintf("%s%s%s", command, url, p + strlen(HELP_BROWSER_PARAM));
        free(old_command);
    }
    system(command);
    free(command);
}

void help_about(void)
{
    GtkWidget*  dialog;
    GtkWidget*  ok_button;
    GtkWidget*  vbox;
    GtkWidget*  hbox;
    GtkWidget*  alignment;
    GtkWidget*  pixmap;

    /* Set up the dialog window */
    dialog = gtk_dialog_new();
    gtk_window_set_title(GTK_WINDOW(dialog), "About " TITLE);
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);

    /* Add the banner */
    pixmap = load_pixmap_from_xpm_file(DATADIR "/graphics/banner.xpm");
    if (pixmap)
        gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), pixmap, FALSE, FALSE, 5);

    hbox = gtk_hbox_new(FALSE, 0);

    /* Add the logo */
    pixmap = load_pixmap_from_xpm_file(DATADIR "/graphics/logo.xpm");
    if (pixmap)
        gtk_box_pack_start(GTK_BOX(hbox), pixmap, FALSE, FALSE, 5);

    /* Add the text */
    vbox = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), gtk_label_new(TITLE " Version " VERSION), FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), gtk_label_new("Suzanne Skinner"), FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), gtk_label_new("Copyright 2003"), FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), gtk_label_new("GNU General Public License"), FALSE, FALSE,
                       0);
    CENTER_ALIGN(vbox);
    gtk_box_pack_start(GTK_BOX(hbox), alignment, TRUE, TRUE, 0);

    /* Add the "okay" button */
    ok_button = gtk_button_new_with_label("Okay");
    gtk_signal_connect_object(GTK_OBJECT(ok_button), "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), ok_button, FALSE, FALSE, 0);
    GTK_WIDGET_SET_FLAGS(ok_button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(ok_button);

    /* Show the dialog */
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hbox, FALSE, FALSE, 5);
    gtk_widget_show_all(dialog);
}

/* Bound Function Helpers */

/* A helper for the various scroll functions. x indicates how to scroll along the x axis (-1 for
 * left, 0 for none, 1 for right), y indicates the y axis scrolling, and by_page indicates whether
 * to scroll by step or by page. This function just adjusts the scrollbar(s), and lets the
 * scrollbar event handlers take care of the rest.
 */
void position_scroll_generic(int32 x, int32 y, boolean by_page)
{
    GtkAdjustment*  sb_state;
    int32           multiplier;

    if (x) {
        sb_state = gtk_range_get_adjustment(GTK_RANGE(hscrollbar));
        multiplier = (by_page ? sb_state->page_increment : sb_state->step_increment);
        gtk_adjustment_set_value(sb_state, ROUND(sb_state->value) + x * multiplier);
    }
    if (y) {
        sb_state = gtk_range_get_adjustment(GTK_RANGE(vscrollbar));
        multiplier = (by_page ? sb_state->page_increment : sb_state->step_increment);
        gtk_adjustment_set_value(sb_state, ROUND(sb_state->value) + y * multiplier);
    }
}

/* Called when the user selects a file from a file open dialog */
void file_open_ok(GtkWidget* widget, GtkFileSelection* file_selector)
{
    char*  path;

    path = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector));
    set_new_current_dir(path);
    if (attempt_load_pattern(path)) {
        sidebar_unselect();
        gtk_widget_destroy(GTK_WIDGET(file_selector));
    }
}

/* Called when the user selects a file from a "save as" dialog */
void file_save_as_ok(GtkWidget* widget, GtkFileSelection* file_selector)
{
    struct stat  statbuf;
    char*  path;

    path = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector));
    if (stat(path, &statbuf) == 0)   /* file exists */
        file_save_as_confirm_overwrite(widget, file_selector);
    else {
        gtk_object_set_user_data(GTK_OBJECT(file_selector), NULL);
        file_save_as_confirmed(widget, file_selector);
    }
}

/* Called to generate an "are you sure" dialog when the user is about to overwrite
 * an existing file.
 */
void file_save_as_confirm_overwrite(GtkWidget* widget, GtkFileSelection* file_selector) {
    GtkWidget*  dialog;
    GtkWidget*  label;
    GtkWidget*  button;
    GtkWidget*  alignment;
    char*       path;
    char*       message;

    dialog = gtk_dialog_new();
    gtk_window_set_title(GTK_WINDOW(dialog), "File Exists");
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
    gtk_object_set_user_data(GTK_OBJECT(file_selector), dialog);
    gtk_container_set_border_width(GTK_CONTAINER(dialog), 0);
    gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 5);

    /* Set up the confirmation question */
    path = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector));
    message = dsprintf("%s exists, overwrite?", path);
    label = gtk_label_new(message);
    free(message);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), label);

    /* Set up the "okay" and "cancel" buttons */
    button = gtk_button_new_with_label("Okay");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(file_save_as_confirmed),
                       file_selector);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(button);
    button = gtk_button_new_with_label("Cancel");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
    CENTER_ALIGN(button);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), alignment, FALSE, FALSE, 0);

    /* Show the dialog */
    gtk_widget_show_all(dialog);
}

/* Called when it's time to actually save to a file--any user confirmation is done. */
void file_save_as_confirmed(GtkWidget* widget, GtkFileSelection* file_selector)
{
    GtkWidget*  confirm_dialog;
    char*       path;

    confirm_dialog = gtk_object_get_user_data(GTK_OBJECT(file_selector));
    if (confirm_dialog)
        gtk_widget_destroy(confirm_dialog);
    path = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector));
    set_new_current_dir(path);
    if (attempt_save_pattern(path)) {
        sidebar_unselect();
        gtk_widget_destroy(GTK_WIDGET(file_selector));
    }
}

/* Called when the user clicks "Okay" from the pattern description dialog */
void file_description_ok(GtkWidget* widget, desc_dialog_info* info)
{
    if (file_description_apply(widget, info))
        gtk_widget_destroy(info->dialog);
}

/* Called when the user clicks "Apply" from the pattern description dialog. Returns TRUE
 * if the description change was successful.
 */
boolean file_description_apply(GtkWidget* widget, desc_dialog_info* info)
{
    static char  buf[DESCRIPTION_MAX_LINES][DESCRIPTION_MAX_COLS+1];
    int32    num_lines = 0;
    boolean  failed = FALSE;
    char*    desc;
    char*    pos;
    char*    newline_pos;
    int32    i;

    desc = gtk_editable_get_chars(GTK_EDITABLE(info->textbox), 0, -1);
    pos = desc;
    while (pos && *pos && !failed) {
        if (num_lines == DESCRIPTION_MAX_LINES)
            failed = TRUE;
        else {
            newline_pos = strchr(pos, '\n');
            if (newline_pos)
                *newline_pos = '\0';
            if (strlen(pos) > DESCRIPTION_MAX_COLS)
                failed = TRUE;
            else {
                strcpy(buf[num_lines++], pos);
                pos = (newline_pos ? newline_pos+1 : NULL);
            }
        }
    }

    free(desc);
    if (failed) {
        error_dialog("The description for a .lif file must be no more\n"
                     "than %d lines long and %d characters wide.\n"
                     "Please fix and retry.",
                     DESCRIPTION_MAX_LINES, DESCRIPTION_MAX_COLS);
        return FALSE;
    } else {
        for (i=0; i < num_lines; i++)
            strcpy(pattern_description[i], buf[i]);
        desc_num_lines = num_lines;
        return TRUE;
    }
}

/* Called when the description dialog is closed */
void file_description_destroy(GtkWidget* widget, desc_dialog_info* info)
{
    free(info);
}

/* Called when the user clicks "Okay" in the preferences dialog */
void edit_preferences_ok(GtkWidget* widget, prefs_dialog_info* info)
{
    edit_preferences_apply(widget, info);
    gtk_widget_destroy(info->dialog);
}

/* Called when the user clicks "Apply" in the preferences dialog */
void edit_preferences_apply(GtkWidget* widget, prefs_dialog_info* info)
{
    static int32 powers_of_2[] = {1, 2, 4, 8, 16, 32, 64, 128, 256, 0};
    GtkAdjustment*  slider_state;
    char*           color_str;
    int32           i;

    /* Apply "File" preferences */
    free(patterns_dir);
    patterns_dir = safe_strdup(gtk_entry_get_text(GTK_ENTRY(info->patterns_dir_entry)));
    if (patterns_dir[strlen(patterns_dir)-1] != '/')
        dstrcat(&patterns_dir, "/");
    update_pattern_list();

    /* Apply General "View" preferences */
    init_window_width =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->init_window_width_spinbox));
    init_window_height =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->init_window_height_spinbox));
    default_zoom = atoi(gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(info->default_zoom_combo)->entry)));
    if (default_zoom < MIN_ZOOM)
        default_zoom = MIN_ZOOM;
    else if (default_zoom > MAX_ZOOM)
        default_zoom = MAX_ZOOM;
    else {
        for (i=0; powers_of_2[i]; i++) {
            if (powers_of_2[i] >= default_zoom) {
                default_zoom = powers_of_2[i];
                break;
            }
        }
    }
    default_show_toolbar = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(info->default_show_toolbar_toggle));
    default_show_sidebar = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(info->default_show_sidebar_toggle));
    default_show_scrollbars = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(info->default_show_scrollbars_toggle));

    /* Apply Color "View" preferences */
    for (i=0; i < NUM_COLORS; i++) {
        color_str = gtk_entry_get_text(GTK_ENTRY(info->color_entries[i]));
        if (color_str[0] == '#')
            color_str++;
        colors[i] = MIN(0xFFFFFF, strtoul(color_str, NULL, 16));
    }
    life_pixmap_colormap = gdk_rgb_cmap_new(colors, NUM_COLORS);

    /* Apply "Run" preferences */
    default_speed = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->default_speed_spinbox));
    speed_slider_min =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->speed_slider_min_spinbox));;
    if (speed_slider_min > default_speed)
        speed_slider_min = default_speed;
    speed_slider_max =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->speed_slider_max_spinbox));;
    if (speed_slider_max < default_speed)
        speed_slider_max = default_speed;
    set_speed_label_size(speed_slider_max);
    speed_increment =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->speed_increment_spinbox));;
    slider_state = gtk_range_get_adjustment(GTK_RANGE(speed_slider));
    slider_state->lower = speed_slider_min;
    slider_state->upper = speed_slider_max;
    slider_state->page_increment = speed_increment;
    gtk_adjustment_changed(slider_state);
    default_skip_frames = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(info->default_skip_toggle));

    /* Apply "Help Browser" preferences */
    free(help_browser);
    help_browser = safe_strdup(gtk_entry_get_text(GTK_ENTRY(info->help_browser_entry)));

    save_preferences();
    full_canvas_redraw();
}

/* Called when the preferences dialog is closed */
void edit_preferences_destroy(GtkWidget* widget, prefs_dialog_info* info)
{
    int32  i;

    for (i=0; i < NUM_COLORS; i++)
        free(info->color_dialog_infos[i]);
    free(info);
}

/* Called when the user clicks "Okay" from the set speed dialog.
 */
void run_speed_ok(GtkWidget* widget, speed_dialog_info* info)
{
    run_speed_apply(widget, info);
    gtk_widget_destroy(info->dialog);
}

/* called when the "Apply" button is clicked in the set speed dialog */
void run_speed_apply(GtkWidget* widget, speed_dialog_info* info)
{
    set_speed(gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->speed_spinbox)));
    skip_frames = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(info->skip_toggle));
}

/* Called when the speed dialog is closed */
void run_speed_destroy(GtkWidget* widget, speed_dialog_info* info)
{
    free(info);
}

/* Called when the user clicks the "..." button to select a new patterns dir in preferences */
void prefs_select_patterns_dir(GtkWidget* widget, prefs_dialog_info* info)
{
    GtkWidget*  file_selector;
    char*       dir;

    /* Setup the file selector */
    file_selector = gtk_file_selection_new("Choose Patterns Directory");
    gtk_window_set_modal(GTK_WINDOW(file_selector), TRUE);
    gtk_window_set_position(GTK_WINDOW(file_selector), GTK_WIN_POS_CENTER);
    gtk_object_set_user_data(GTK_OBJECT(file_selector), info);
    dir = safe_strdup(gtk_entry_get_text(GTK_ENTRY(info->patterns_dir_entry)));
    if (dir[strlen(dir)-1] != '/')
        dstrcat(&dir, "/");
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_selector), dir);
    free(dir);

    /* Link the buttons to handlers */
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->ok_button), "clicked",
                       GTK_SIGNAL_FUNC(prefs_select_patterns_dir_ok), file_selector);
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->cancel_button),
                              "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(file_selector));

    /* Show the dialog */
    gtk_widget_show(file_selector);
}

/* Called when the user finishes selecting a new patterns dir */
void prefs_select_patterns_dir_ok(GtkWidget* widget, GtkFileSelection* file_selector)
{
    prefs_dialog_info*  info;

    info = gtk_object_get_user_data(GTK_OBJECT(file_selector));
    gtk_entry_set_text(GTK_ENTRY(info->patterns_dir_entry),
                       gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector)));
    gtk_widget_destroy(GTK_WIDGET(file_selector));
}

/* Called when the user clicks the "default" button for patterns directory in preferences */
void prefs_default_patterns_dir(GtkWidget* widget, prefs_dialog_info* info)
{
    gtk_entry_set_text(GTK_ENTRY(info->patterns_dir_entry), DEFAULT_PATTERNS_DIR);
}

/* Called when the user clicks a "select" button to change a color in preferences */
void prefs_select_color(GtkWidget* widget, color_dialog_info* info)
{
    GtkWidget*  color_dialog;
    double      color_array[4];
    int32       i;

    /* Setup the color selector */
    color_dialog = gtk_color_selection_dialog_new(info->dialog_title);
    gtk_window_set_modal(GTK_WINDOW(color_dialog), TRUE);
    gtk_window_set_position(GTK_WINDOW(color_dialog), GTK_WIN_POS_CENTER);
    gtk_object_set_user_data(GTK_OBJECT(color_dialog), info);
    color_array[0] = (double)(info->cur_value >> 16)         / 255.0;
    color_array[1] = (double)((info->cur_value >> 8) & 0xFF) / 255.0;
    color_array[2] = (double)(info->cur_value & 0xFF)        / 255.0;
    for (i=0; i < 2; i++)
        gtk_color_selection_set_color(
            GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(color_dialog)->colorsel), color_array);

    /* Link the buttons to handlers */
    gtk_signal_connect(GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(color_dialog)->ok_button), "clicked",
                       GTK_SIGNAL_FUNC(prefs_select_color_ok), color_dialog);
    gtk_signal_connect_object(GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(color_dialog)->cancel_button),
                              "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(color_dialog));

    /* Show the dialog */
    gtk_widget_show(color_dialog);
}

/* Called when the user finishes selecting a new color in preferences */
void prefs_select_color_ok(GtkWidget* widget, GtkColorSelectionDialog* color_dialog)
{
    color_dialog_info*  info;
    double              color_array[4];
    char*               color_str;

    info = gtk_object_get_user_data(GTK_OBJECT(color_dialog));
    gtk_color_selection_get_color(
        GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(color_dialog)->colorsel), color_array);
    info->cur_value = ((int)(color_array[0] * 255.0) << 16) |
                      ((int)(color_array[1] * 255.0) << 8)  |
                      ((int)(color_array[2] * 255.0));
    color_str = dsprintf("#%06X", info->cur_value);
    gtk_entry_set_text(GTK_ENTRY(info->prefs_entry), color_str);
    free(color_str);
    gtk_widget_destroy(GTK_WIDGET(color_dialog));
}

/* Called when the user clicks "default" to reset a color in preferences */
void prefs_default_color(GtkWidget* widget, color_dialog_info* info)
{
    char*  color_str;

    info->cur_value = info->default_value;
    color_str = dsprintf("#%06X", info->cur_value);
    gtk_entry_set_text(GTK_ENTRY(info->prefs_entry), color_str);
    free(color_str);
}

/* Called when the user clicks "default" to reset the help browser in preferences */
void prefs_default_help_browser(GtkWidget* widget, prefs_dialog_info* info)
{
    gtk_entry_set_text(GTK_ENTRY(info->help_browser_entry), DEFAULT_HELP_BROWSER);
}

/* Miscellaneous Functions */

/* Add the current file to the top of the recent files list, and update the File menu accordingly.
 * If it is already in the list, move it to the top.
 */
void add_to_recent_files(void)
{
    char     full_path[PATH_MAX];
    char*    filename;
    char*    slash_pos;
    char*    label_text;
    boolean  maxed_out = FALSE;
    int32    num_recent;
    int32    old_pos;
    int32    start;
    int32    i;

    /* Get the full path of the current file */
    if (!realpath(pattern_path, full_path)) {
        sys_warn("can't get realpath of %s", pattern_path);
        strcpy(full_path, pattern_path);
    }

    /* Is it already at the top of recent files? */
    if (recent_files[0].full_path && STR_EQUAL(full_path, recent_files[0].full_path))
        return;

    /* Determine filename from full_path */
    slash_pos = strrchr(full_path, '/');
    filename = safe_strdup(slash_pos ? slash_pos+1 : full_path);

    /* Determine the current # of recent files, and the position of this one if it's already in
     * the list.
     */
    for (i=0, old_pos=0; i < MAX_RECENT_FILES && recent_files[i].full_path; i++) {
        if (STR_EQUAL(recent_files[i].full_path, full_path))
            old_pos = i;
    }
    num_recent = i;

    /* If the recent files list is at maximum and this is a new entry, drop off the last item */
    if (num_recent == MAX_RECENT_FILES && !old_pos) {
        maxed_out = TRUE;
        num_recent--;
    }

    /* Make room for the new item at the top of the list */
    start = (old_pos ? old_pos : num_recent);
    free(recent_files[start].full_path);
    free(recent_files[start].filename);
    for (i=start; i > 0; i--) {
        recent_files[i].full_path = recent_files[i-1].full_path;
        recent_files[i].filename  = recent_files[i-1].filename;
        label_text = dsprintf("%d. %s", i+1, recent_files[i].filename);
        gtk_label_set_text(GTK_LABEL(recent_files[i].label), label_text);
        free(label_text);
    }

    /* Put the new item at the top */
    recent_files[0].full_path = safe_strdup(full_path);
    recent_files[0].filename  = filename;
    label_text = dsprintf("1. %s", filename);
    gtk_label_set_text(GTK_LABEL(recent_files[i].label), label_text);
    free(label_text);

    /* If we've increased the length of the list, show the menu item which was hidden before */
    if (!maxed_out && !old_pos)
        gtk_widget_show(recent_files[num_recent].menu_item);
    /* If the list was empty before this, show the menu separator which was hidden before */
    if (!num_recent)
        gtk_widget_show(recent_files_separator);
}

/* Attempt to load a pattern file from the given path, displaying an error dialog and returning
 * FALSE on failure, otherwise returning TRUE. If the original path cannot be loaded, also try
 * with .DEFAULT_EXTENSION appended. If the load was successful, this function will stop any
 * running pattern, update the tick label, and recenter and redraw the canvas.
 */
boolean attempt_load_pattern(const char* path)
{
    char*             path2;
    boolean           with_extension = FALSE;
    load_result_type  load_result;

    load_result = load_pattern(path);
    if (load_result != LOAD_SUCCESS) {
        path2 = dsprintf("%s.%s", path, DEFAULT_EXTENSION);
        if (load_pattern(path2) == LOAD_SUCCESS) {
            path = path2;
            with_extension = TRUE;
            load_result = LOAD_SUCCESS;
        }
        else
            free(path2);
    }

    if (load_result == LOAD_SUCCESS) {
        if (!pattern_path || !STR_EQUAL(path, pattern_path)) {
            free(pattern_path);
            pattern_path = (with_extension ? (char*)path : safe_strdup(path));
            add_to_recent_files();
            put_filename_in_window_title();
        }
        if (pattern_running)
            start_stop();
        update_tick_label();
        update_population_label();
        position_center();   /* will trigger a canvas redraw */
        return TRUE;
    }

    if (load_result == LOAD_NOT_FOUND)
        error_dialog("Open failed: \"%s\" not found", path);
    else if (load_result == LOAD_SYS_ERROR)
        error_dialog("Open failed: %s", strerror(errno));
    else
        error_dialog("Open failed: invalid format in \"%s\"", path);
    return FALSE;
}

/* Attempt to save a pattern file to the given path, displaying an error dialog and returning
 * FALSE on failure, otherwise returning TRUE.
 */
boolean attempt_save_pattern(const char* path)
{
    if (save_pattern(path)) {
        if (path != pattern_path) {
            free(pattern_path);
            pattern_path = safe_strdup(path);
            add_to_recent_files();
            put_filename_in_window_title();
        }
        return TRUE;
    }
    else {
        error_dialog("Save failed: %s", strerror(errno));
        return FALSE;
    }
}

/* Set current_dir (default load directory) to the parent directory of the given file.
 */
void set_new_current_dir(const char* filepath)
{
    char*  parent_dir;
    char*  slash_pos;

    parent_dir = safe_strdup(filepath);
    slash_pos = strrchr(parent_dir, '/');
    if (!slash_pos) {
        free(parent_dir);
        return;
    }
    *(slash_pos+1) = '\0';
    free(current_dir);
    current_dir = parent_dir;
}

/* Set the pattern run speed, by some means other than the speed slider. This function will
 * readjust the slider, thus triggering an event, so handle_speed_slider_change will take care of
 * everything else.
 */
void set_speed(int32 new_speed)
{
    GtkAdjustment*  slider_state;

    speed = new_speed;
    if (speed < 1)
        speed = 1;

    slider_state = gtk_range_get_adjustment(GTK_RANGE(speed_slider));
    if (speed > slider_state->upper) {
        slider_state->upper = speed;
        gtk_adjustment_changed(slider_state);
        set_speed_label_size(speed);
    }
    gtk_adjustment_set_value(slider_state, speed);
}

/* Deselect any selected item in the sidebar.
 */
void sidebar_unselect(void)
{
    gtk_clist_unselect_all(GTK_CLIST(pattern_list));
    GTK_CLIST(pattern_list)->focus_row = -1;
    gtk_widget_queue_draw(pattern_list);
}

/* Toggle pattern_running, and update menus, toolbar, etc. accordingly, but do *not* update the
 * canvas or the tick label.
 */
void start_stop(void)
{
    const char*  new_text;
    GtkWidget*   new_icon;

    pattern_running = !pattern_running;
    if (pattern_running) {
        skipped_frames = 0;
        start_tick = tick;
        start_time = get_time_milliseconds();
        new_text = "Stop";
        new_icon = stop_pixmap;
    } else {
        skipped_frames = 0;
        new_text = "Start";
        new_icon = start_pixmap;
    }

    gtk_label_set_text(GTK_LABEL(start_stop_menu_label), new_text);
    gtk_container_add(GTK_CONTAINER(start_stop_toolbar_button), new_icon);
    gtk_widget_show(new_icon);
}

/* Utility Functions */

/* Get the current time in milliseconds, as a 64-bit uint
 */
uint64 get_time_milliseconds(void)
{
    struct timeval   t;
    struct timezone  tz;

    gettimeofday(&t, &tz);
    return (uint64)(t.tv_sec) * 1000ULL + (uint64)(t.tv_usec) / 1000ULL;
}

/* Return true if the given string is composed entirely of numeric digits
 */
boolean is_numeric(const char* str)
{
    const char*  p;

    for (p=str; *p; p++) {
        if (!isdigit(*p))
            return FALSE;
    }
    return TRUE;
}

/* Load a GTK pixmap from an xpm file, returning NULL on failure.
 */
GtkWidget* load_pixmap_from_xpm_file(const char* path)
{
    GdkPixmap*  gdk_pixmap;

    gdk_pixmap = gdk_pixmap_create_from_xpm(window->window, NULL,
                                            &(window->style->bg[GTK_STATE_NORMAL]), path);
    if (!gdk_pixmap)
        return NULL;
    return gtk_pixmap_new(gdk_pixmap, NULL);
}

/* Load a GTK Pixmap from an array of RGBA pixel values, alpha blending onto the given
 * background color.
 */
GtkWidget* load_pixmap_from_rgba_array(uint8* pixels, int32 width, int32 height, GdkColor* bg)
{
    GdkPixmap*  gdk_pixmap;
    uint8       bg_rgb[3];
    uint8*      rgb_pixels;
    int32       i;

    bg_rgb[0] = bg->red   / 257;
    bg_rgb[1] = bg->green / 257;
    bg_rgb[2] = bg->blue  / 257;
    rgb_pixels = safe_malloc(width * height * 3);
    for (i=0; i < width*height; i++) {
        memcpy(&rgb_pixels[i*3], bg_rgb, 3);
        alpha_blend(&rgb_pixels[i*3], &pixels[i*4], pixels[i*4+3]);
    }
    gdk_pixmap = gdk_pixmap_new(window->window, width, height, -1);
    gdk_draw_rgb_image(gdk_pixmap, window->style->fg_gc[GTK_WIDGET_STATE(window)], 0, 0, width,
                       height, GDK_RGB_DITHER_NONE, rgb_pixels, width*3);
    free(rgb_pixels);
    return gtk_pixmap_new(gdk_pixmap, NULL);
}

/* Blend the pixel rgb2 (an array of 3 unsigned chars represented red, green,
 * blue) onto the pixel rgb1, where rgb2 has the given opacity (0-255).
 */
void alpha_blend(uint8* rgb1, const uint8* rgb2, int32 opacity)
{
    *rgb1 = ((int32)(*rgb1++) * (255 - opacity) + (int32)(*rgb2++) * opacity) / 255;
    *rgb1 = ((int32)(*rgb1++) * (255 - opacity) + (int32)(*rgb2++) * opacity) / 255;
    *rgb1 = ((int32)(*rgb1++) * (255 - opacity) + (int32)(*rgb2++) * opacity) / 255;
}
